"""Protein libraries.

This module provides the RecombinantLibrary class that represents a library of
chimeric proteins from a given alignment of parent amino acids sequences. Each
library is defined by a series of breakpoints, which are the indices where the
parent alignment is recombined into chimeras.
"""

# from dataclasses import dataclass, field
from functools import cached_property
from itertools import combinations

from ggrecomb import ParentAlignment


# @dataclass(repr=False)
class RecombinantLibrary:
    def __init__(
        self,
        energy: float,
        breakpoints: dict[int, list[tuple[int, str]]],
        parent_alignment: ParentAlignment,
        bp_to_group: dict[int, int] = None,
        group_bp_cache: dict[tuple[int, ...], float] = None,
    ):
        self.energy = energy
        self.breakpoints = breakpoints
        self.parent_alignment = parent_alignment

        self.mutation_rate = self.calc_average_m(
            parent_alignment,
            bp_to_group,
            group_bp_cache
        )

    @cached_property
    def min_block_len(self):
        bps = sorted(self.breakpoints)
        return min(bp2 - bp1 for bp1, bp2 in zip(bps, bps[1:]))

    @cached_property
    def max_block_len(self):
        bps = sorted(self.breakpoints)
        return max(bp2 - bp1 for bp1, bp2 in zip(bps, bps[1:]))

    def expand(self, e_diff: float, new_bp: dict[int, list[tuple[int, str]]]):
        new_e = self.energy + e_diff
        new_bps = self.breakpoints | new_bp
        return type(self)(new_e, new_bps)

    def __repr__(self):
        return f'Library({self.energy}, {list(self.breakpoints.keys())})'

    @property
    def average_m(self):
        return self._m

    '''
    def calc_average_m_naive(self, pa: ParentAlignment, bp_to_group,
                             group_bp_cache):
        """Calculate the average number of mutations in library.

        Naive way to calculate <M>. Currently kept for testing purposes, may be
        deleted later for simplicity.
        """
        def calc_muts(seq1, seq2):
            return sum(1 for a1, a2 in zip(seq1, seq2) if a1 != a2)

        bps = tuple(self.breakpoints)

        # ADDITION
        bp_groups = tuple(bp_to_group[bp] for bp in bps)
        if (m := group_bp_cache.get(bp_groups)) is not None:
            self._m = m
            return m
        # END ADDITION

        p_seqs = [str(sr.seq) for sr in pa.aligned_sequences]

        p = len(p_seqs)  # number of parents
        N = len(p_seqs[0])  # number of amino acids in alignment

        # bps must be (0, ..., N).
        if bps[0] != 0:
            bps = (0,) + bps
        if bps[-1] != N:
            bps += (N,)

        n = len(bps) - 2  # number of crossovers. number of blocks is n+1

        # Construct all library chimeras to calculate average mutations.
        total_muts = 0
        for i, block_parents in enumerate(product(p_seqs, repeat=n+1)):
            # Construct chimera from parent blocks.
            block_seqs = [blkpar[start: end] for blkpar, start, end
                          in zip(block_parents, bps, bps[1:])]
            chimera = ''.join(block_seqs)

            # Chimera m is the number of mutations from the closest parent.
            total_muts += min(calc_muts(chimera, parent) for parent in p_seqs)

        m = total_muts / p**(n+1)

        # ADDITION
        # global group_bp_map
        group_bp_cache[bp_groups] = m
        # END ADDITION

        self._m = m
        return m
    '''

    def calc_average_m(self, pa: ParentAlignment, bp_to_group, group_bp_cache):
        """Calculate the average number of mutations in library."""
        def sequence_mutations(seq1, seq2):
            """Hamming distance between seq1 and seq2."""
            assert len(seq1) == len(seq2)
            return sum(1 for a1, a2 in zip(seq1, seq2) if a1 != a2)

        print(bp_to_group)
        print(group_bp_cache)
        if bp_to_group is not None and group_bp_cache is not None:
            bp_groups = tuple(bp_to_group[bp] for bp in self.breakpoints)
            if (M := group_bp_cache.get(bp_groups)) is not None:
                self._m = M
                return M

        seqs = [str(sr.seq) for sr in pa.aligned_sequences]
        blmuts = []
        bps = sorted(self.breakpoints)
        for bp1, bp2 in zip(bps, bps[1:]):
            muts = {i: {i: 0} for i, _ in enumerate(seqs)}
            for (i, s1), (j, s2) in combinations(enumerate(seqs), 2):
                s1_slice = s1[bp1:bp2]
                s2_slice = s2[bp1:bp2]
                m = sequence_mutations(s1_slice, s2_slice)
                muts[i][j] = m
                muts[j][i] = m
            blmuts.append(muts)

        # Calculate parameters for mutation calculation.
        num_blocks = len(self.breakpoints) - 1
        num_parents = len(seqs)

        # Perform depth first traversal over blocks to calculate M_sum. This
        # is faster than itertools.product because it avoids redundant
        # summation.
        stack = [(1, [blmuts[0][p1][p2] for p2 in range(num_parents)])
                 for p1 in range(num_parents)]
        M_sum = 0
        while stack:
            i, par_muts = stack.pop()
            par_iter = list(zip(par_muts, range(num_parents)))
            block_muts = blmuts[i]
            for p1 in range(num_parents):
                p_muts = block_muts[p1]
                new_part_muts = [m + p_muts[p2] for m, p2 in par_iter]
                if i + 1 == num_blocks:
                    M_sum += min(new_part_muts)
                else:
                    stack.append((i + 1, new_part_muts))

        M = M_sum / num_parents**num_blocks
        self._m = M

        if bp_to_group is not None and group_bp_cache is not None:
            group_bp_cache[bp_groups] = M

        return M



from itertools import product

import numpy as np

from ggrecomb import ParentAlignment


class EnergyFunction:
    # TODO: Implement general energy function.
    def __init__(self, pa: ParentAlignment):
        raise NotImplementedError


class SCHEMA(EnergyFunction):
    """SCHEMA energy.

    Given a PDB structure and a multiple sequence alignment, the SCHEMA energy
    of a recombinant sequence is the number of pairwise interactions broken by
    recombination. Expressed as an equation (Endelman et al. 2004 eq. 7):

        E = sum_i sum_{j>i} C_{i,j} D_{i,j}

        where:
            C_{i,j} = 1 if residues i and j are within 4.5 Angstroms, else 0.
            D_{i,j} = 0 if the amino acids at positions i and j in the
                recombinant sequence are also present together in any parent,
                else 0.

    Example: If "AEH" and "GQH" are aligned parents with contacts (0, 1) and
        (1, 2), chimera "AQH" has a SCHEMA energy of 1.0 because residues 'A'
        and 'Q' are not found in any parents at positions 0 and 1,
        respectively, and there is a contact at these positions. Although
        residues 1 and 2 are contacting, 'Q' and 'H' are found together at
        these positions in the second parent, so D_{0,1} = 0 and doesn't
        contribute to the SCHEMA energy.

    See Voigt et al. 2002 and Silberg et al. 2004 for more information.

    Public methods:
        block_energy: Average SCHEMA energy of a section of a parent multiple
            sequence alignment.
        increment_block_energy: Average SCHEMA energy difference between a
            block and a block decreased by one residue.
    """
    def __init__(self, pa: ParentAlignment):
        """Generate the SCHEMA energy matrix for block calculation.

        E_matrix is a square matrix with the same size as the length of
        alignment. E_matrix_{r,j} is Endelman et al. 2004 eq. 6. with the r and
        t sums removed and evaluated for average SCHEMA energy for a specific r
        and t. Then eq. 6 can be evaluated for a given block by summing over
        E_matrix_{r,t}.

        Args:
            pa: Parent sequence alignment. Must have pdb_structure attribute.
        """
        alignment = list(zip(*pa.aligned_sequences))
        self.E_matrix = np.zeros((len(alignment), len(alignment)))

        # Only need to evaluate at contacts because otherwise C_{i,j} = 0.
        for i, j in pa.pdb_structure.contacts:
            # Get the ordered parent amino acids at i and j.
            AAs_i, AAs_j = alignment[i], alignment[j]

            # Parent amino acid pair might be redundant.
            parental = set(zip(AAs_i, AAs_j))

            # For each possible recombinant calculate D_ij. E_matrix entry
            # is this sum over the number of possible recombinants.
            broken = sum(1 for combo in product(AAs_i, AAs_j) if combo
                         not in parental)
            self.E_matrix[i, j] = broken / (len(AAs_i)**2)

    def block_energy(self, start: int, end: int) -> float:
        """Calculate the average SCHEMA energy of a block.

        This is Endelman et al. 2004 eq 6.

        Args:
            start: Index of the first residue in the block.
            end: Index of the first residue in the next block. Last residue in
                the current block is at end-1.

        Returns:
            Average SCHEMA energy of block [start, end].
        """
        energy = self.E_matrix[start:end, end:].sum()
        return energy

    def increment_block_energy(self, start: int, new_end: int) -> float:
        """Difference in average SCHEMA energy when block size is incremented.

        This is equivalent to
            block_energy(start, new_end) - block_energy(start, new_end-1).

        This function is faster than separate block_energy calculations.
        Therefore, you can calculate the energy for the first node in the
        graph using block_energy, then calculate the subsequent node energies
        using this function.

        Args:
            start: Index of the first residue in the block.
            end: Index of the first residue in the next block, after the
                current block size is incremented.

        Returns:
            Difference in average SCHEMA energy between block [start, new_end]
                and block [start, new_end-1].
        """
        old_end = new_end - 1
        neg = self.E_matrix[start:old_end, old_end].sum()
        pos = self.E_matrix[old_end, new_end:].sum()
        return pos - neg



""" Golden Gate SCHEMA-RASPP algorithm.

Run syntax: "python step2.py <num_bl> <minBL> <maxBL> <alignment_fn>
    <contacts_fn> <libraries_fn>"
Example: "python step2.py 8 50 300 bgl3_AA.fasta bgl3_contacts.json
    bgl3_libraries.json"

Command Line Args:
    num_bl: number of blocks in libraries
    minBL: minimum block length in libraries
    maxBL: maximum blocki length in libraries
    start_oh_seq: sequence of start vector overhang
    temp_start_oh_pos: position of start vector overhang relative to first base
        of parent sequence. Can range from -4 to -1, inclusive. Must be
        converted to position relative to imaginary codon at position -1 (i.e.
        add 3) to stay consistent with breakpoint/overhang positioning
        convention
    end_oh_seq: sequence of end vector overhang
    end_oh_pos: position of end vector overhang first base pair relative to
        last codon in parent sequences. Can range from 0 to 3, inclusive
    alignment_fn: name of amino acid alignment file, from step 1
    contacts_fn: name of contacts file, from step 1
    libraries_fn: output filename

Output:
    libraries dictionary in <libraries_fn>, used in step 3

You can also replace all instances of "sys.argv" in the code with the input
filenames directly, then run "python step2.py".

This script generates the graph from Endelman JB et al. "Site-directed protein
recombination as a shortest-path problem." (2004), but uses Dijkstra's
algorithm to solve the shortest-path problem. While this is technically less
efficient, the reduction in number of breakpoints caused by the Golden Gate
criteria makes enumeration of nearly all paths tractable.

It's recommended to run this script with "nohup <command> &> log.txt &" because
it may take awhile. You can then leave the server and the script will keep
running. Output will be redirected to log.txt.

BioPython is a required package. The Romero Lab group server has it installed.
"""

import json
import sys

from Bio import SeqIO

from tools import step2_tools

# Read inputs from sys.argv. Can also replace sys.argv below with strings/ints.
num_bl = int(sys.argv[1])  # number of blocks in libraries
minBL = int(sys.argv[2])  # minimum block length in libraries
maxBL = int(sys.argv[3])  # maximum blocki length in libraries
start_oh_seq = sys.argv[4]  # sequence of start vector overhang
temp_start_oh_pos = int(sys.argv[5])  # position of start vector overhang,
# (must be converted to position relative to 1 imaginary codon back)
start_oh_pos = temp_start_oh_pos + 3
end_oh_seq = sys.argv[6]  # sequence of end vector overhang
end_oh_pos = int(sys.argv[7])  # position of end vector overhang
alignment_fn = sys.argv[8]  # name of alignment file from Step1
contacts_fn = sys.argv[9]  # name of contacts file from Step1
libraries_fn = sys.argv[10]  # output filename

vector_overhangs = ((start_oh_pos, start_oh_seq),
                    (end_oh_pos, end_oh_seq))

print(f'Running SCHEMA-RASPP with the following parameters:', flush=True)
print(f'{num_bl} blocks, minimum block length: {minBL}, maximum block '
      f'length: {maxBL}', flush=True)
print(f'Loaded alignment file: "{alignment_fn}"')
print(f'Loaded contacts file: "{contacts_fn}"')

# Load muscle output fasta file. AA_alignment is a list of tuples where each
# tuple contains the residues at the tuple index position in the MSA.
alignment_SRs = list(SeqIO.parse(alignment_fn, 'fasta'))
AA_seqs = [str(i.seq) for i in alignment_SRs]
AA_alignment = list(zip(*AA_seqs))  # list of AAs at each position

# Load contact dictionary. Dictionary values are 1 to correspond to unweighted
# contact behavior. Change dictionary values to weight contacts.
with open(contacts_fn, 'r') as f:
    contacts = json.load(f)
contacts = {tuple(c): 1 for c in contacts}

# A breakpoint specifies the first position of a new block. breakpoints is a
# dictionary with <breakpoint>: <potential GG overhangs> pairs. Note that
# breakpoints are restricted to having 4 possible overhangs to speed up the
# update GG_prob call below. Increasing this number will potentially lead to
# factorial increases in run time for this call. Set to 'None' to avoid this
# restriction and potentially run this step faster.
print('Finding breakpoints', flush=True)
breakpoints = step2_tools.find_GG_breakpoints(AA_alignment, vector_overhangs,
                                              max_num_ohs=4)
print(f'{len(breakpoints)} breakpoints found', flush=True)

# Generate matrix used to calulcate SCHEMA energy.
E_matrix = step2_tools.generate_weighted_E_matrix(AA_alignment, contacts)
print('E matrix generated', flush=True)

# Generate all allowed blocks. blocks is a list of tuples with starting and
# ending breakpoints.
blocks = step2_tools.generate_blocks(breakpoints, minBL, maxBL)
print('Blocks generated', flush=True)

# Generate libraries with the minimum value for a given minimum and maximum
# block length combination. libraries is a dictionary of
# lib_bps: {'E':' <SCHEMA energy>} pairs where lib_bps is a tuple of
# breakpoint indices.
print('Starting shortest path recombination to calculate E', flush=True)
libraries = step2_tools.shortest_path_recombination(num_bl, blocks, E_matrix)
lib_len = len(libraries)
print(f'Shortest path recombination complete: {lib_len} libraries found',
      flush=True)

# Add 'M': <average chimera mutations> pairs to the libraries values.
print('Updating M', flush=True)
step2_tools.update_M(libraries, AA_alignment)
print('M updated', flush=True)

# Add 'GG_prob': <Golden Gate probability> and 'GG_sites': <valid overhangs
# for each lib_bp> pairs the libraries values.
# Requires 'normalized_ligation_counts_18h_37C.p' file to be in code directory.
print('Updating GG probabilities', flush=True)
libraries = step2_tools.update_GG_prob(libraries, breakpoints)
print('GG probabilities updated', flush=True)

# Convert libraries to be json-compatible and save.
libraries = [(k, v) for k, v in libraries.items()]
with open(libraries_fn, 'w') as f:
    json.dump(libraries, f)

print(f'Saved libraries to "{libraries_fn}"', flush=True)



import itertools

from Bio import Seq

# DNA bases and codon->AA translation code.
bases = ['A', 'C', 'G', 'T']
codons = [''.join(cdn) for cdn in itertools.product(bases, repeat=3)]
code = {cdn: str((Seq.Seq(cdn)).translate()) for cdn in codons}
code['---'] = '-'

# Remove rare codons -- we don't want to allow these at the block junctions.
# (Phil Romero found these from various sources online.)
rare = ('ATA', 'AGG', 'TCG', 'ACG', 'AGA', 'GGA', 'CTA', 'GGG', 'CCC', 'CGG',
        'CGA')
(code.pop(r) for r in rare)

# AA->codon reverse translation code.
aminos = set(code.values())
rev_code = {a: tuple([k for k in code.keys() if code[k] == a]) for a in aminos}


class ProgressOutput:
    """ Used to show progress during iterative processes. """
    def __init__(self, total_itrs):
        """ Need total number of iterations that will be run. """
        self.total_itrs = total_itrs
        self.printed_percentage = -1

    def update(self, current_itr):
        """ Called at start of each iteration. """
        current_percentage = round(current_itr * 100 / self.total_itrs)
        if current_percentage != self.printed_percentage:
            print(f'{current_percentage}% done\r', end='', flush=True)
        self.printed_percentage = current_percentage



""" This module contains tools used by the Step2 script. """

import collections
import itertools
import os
import pickle
from random import choice, sample
import re
import sys

import networkx
import numpy as np
# import pandas as pd

from tools import general_tools

# Change these as needed
dirname = os.path.split(os.path.abspath(os.path.realpath(sys.argv[0])))[0]
# lig_counts_fn = dirname + '/tools/normalized_ligation_counts_18h_37C.p'
# normalized_ligation_counts = pickle.load(open(lig_counts_fn, 'rb'))
pd_lig_counts_fn = dirname + '/tools/pd_normalized_ligation_counts_18h_37C.p'
# normalized_ligation_counts = pd.DataFrame(normalized_ligation_counts)
# pickle.dump(normalized_ligation_counts, open(pd_lig_counts_fn, 'wb'))
normalized_ligation_counts = pickle.load(open(pd_lig_counts_fn, 'rb'))

threshold = 245  # ligation count threshold for eliminating bad overhangs

complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'}


def complementary_sequence(sequence):
    """ Returns the DNA complement of sequence. """
    return ''.join(reversed([complement[bp] for bp in sequence]))


def acceptable_overhang(overhang, vector_overhangs):
    """Returns true if overhang is not palindromic, full-GC, below threshold,
    or is not a vector_overhang. """

    # palindromic overhangs
    if overhang == complementary_sequence(overhang):
        return False

    # high GC overhangs
    if 'A' not in overhang and 'T' not in overhang:
        return False

    # overhangs with low rates of correct ligations
    if normalized_ligation_counts[overhang][complementary_sequence(overhang)] \
       < threshold:
        return False

    # vector overhangs
    if overhang in vector_overhangs or complementary_sequence(overhang) \
       in vector_overhangs:
        return False

    return True


all_overhangs = [''.join(oh) for oh in itertools.product(general_tools.bases,
                                                         repeat=4)]


def _pattern_valid(pattern, pos_cdns):
    """ Check if pattern can represent pos_cdns. Helper function for
    get_valid_patterns.

    Args:
        pattern: str of length 3 that represents a codon pattern. Wild card
            base positions are denoted with '.' for re compatibility.
        pos_cdns: List of lists of codons. Each inner list has all valid codons
            for one of the unique residues at the given MSA position.

    Returns:
        True if at least one codon in each pos_cdns element follows pattern.
    """
    for AA_cdns in pos_cdns:
        matches_a_cdn = any(re.fullmatch(pattern, codon) for codon in AA_cdns)
        if not matches_a_cdn:
            return False
    return True


def _get_valid_patterns(pos_cdns, reverse=False):
    """ Gets patterns that describe at least one codon in each pos_cdn element.

    Each pattern is a length 3 str that may use '.' as a wild card.
    Patterns are explored left to right e.g. if 'A..' works then 'AA.', 'AG.'
    are explored, unless reverse=True.

    Args:
        pos_cdns: List of lists of codons. Each inner list has all valid codons
            for one of the unique residues at the given MSA position.
        reverse: Boolean specifying if patterns should be explored right to
            left.

    Returns:
        List of valid patterns.
    """
    # reverse cdns if pattern search is at end of cdns rather than start
    if reverse:
        pos_cdns = [[cdn[::-1] for cdn in AA_cdns] for AA_cdns in pos_cdns]

    # need to iterate through all possible cdn patterns, but can eliminate
    # based on subpattern validity
    valid_patterns = []
    for b1 in general_tools.bases:
        pattern = f'{b1}..'
        if not _pattern_valid(pattern, pos_cdns):
            continue  # No use exploring derivatives
        valid_patterns.append(pattern)
        for b2 in general_tools.bases:
            pattern = f'{b1}{b2}.'
            if not _pattern_valid(pattern, pos_cdns):
                continue  # No use exploring derivatives
            valid_patterns.append(pattern)
            for b3 in general_tools.bases:
                pattern = f'{b1}{b2}{b3}'
                if _pattern_valid(pattern, pos_cdns):
                    valid_patterns.append(pattern)

    # reverse back
    if reverse:
        valid_patterns = [vp[::-1] for vp in valid_patterns]

    return valid_patterns


def _patterns_to_overhangs(AA1_patterns, AA2_patterns, acceptable_overhangs):
    """ Converts patterns from the two codon sets into the possible overhangs.

    Essentially finds combinations of patterns from the pattern lists that are
    length 4 when wildcards are eliminated. These combination are valid Golden
    Gate overhangs. The number of left wildcards is recorded with the overhang
    to represent the position of the overhang relative to the two breakpoint
    codons.

    Args:
        AA1_patterns: List of valid patterns for first residue at breakpoint.
            Wildcard '.' should be at the left of patterns.
        AA2_patterns: List of valid patterns for second residue at breakpoint.
            Wildcard '.' should be at the right of patterns.
        acceptable_overhangs: List of overhangs that pass the criteria in
            acceptable_overhang(). Each overhang is of the form
            (overhang index, overhang sequence).

    Returns:
        List of overhangs that are valid for the given breakpoint.
    """
    # bin patterns by length
    AA1_binned = [[], [], []]
    AA2_binned = [[], [], []]
    for pat in AA1_patterns:
        pat = pat.replace('.', '')
        AA1_binned[len(pat) - 1].append(pat)
    for pat in AA2_patterns:
        pat = pat.replace('.', '')
        AA2_binned[len(pat) - 1].append(pat)
    AA1_binned = reversed(AA1_binned)  # reverse for combination with AA2_pats

    # find any length 4 overhangs (e.g. all len 1 AA1 pats w/ len 3 AA pats)
    # pos is 3 - len(AA1_pat)
    overhangs = []
    for pos, (AA1_pats, AA2_pats) in enumerate(zip(AA1_binned, AA2_binned)):
        if not AA1_pats or not AA2_pats:
            continue
        for pat1, pat2 in itertools.product(AA1_pats, AA2_pats):
            oh = pat1 + pat2
            if oh in acceptable_overhangs:
                overhangs.append((pos, oh))
    return overhangs


def _calculate_GG_prob(overhang_seqs):
    """ Golden Gate ligation probability.

    Calculates the pseudo-probability of correct GG ligation given a list of
    overhang sequences. Ligation data is from Potapov V et al 2018. This isn't
    a real probability but scales from 0 to 1 and is roughly representative
    of the compatibility of the overhangs in the set.

    Args:
        overhang_seqs: List of str representing overhang DNA sequences.

    Returns:
        float probability of GG ligation.
    """
    complements = [complementary_sequence(oh) for oh in overhang_seqs]

    all_overhang_seqs = overhang_seqs + list(reversed(complements))

    # construct matrix of pairwise overhang ligation counts
    '''
    overhang_set_counts = []
    for oh1 in all_overhang_seqs:
        oh_set_counts_row = []
        for oh2 in reversed(all_overhang_seqs):
            oh_set_counts_row.append(normalized_ligation_counts[oh1][oh2])
        overhang_set_counts.append(oh_set_counts_row)
    '''
    rev_seqs = list(reversed(all_overhang_seqs))
    overhang_set_counts = normalized_ligation_counts.loc[all_overhang_seqs,
                                                         rev_seqs]
    np_oh_counts = overhang_set_counts.to_numpy()

    # find product of matrix diagonals each divided by their row sum
    '''
    GG_prob = 1
    for i in range(len(overhang_set_counts)):
        GG_prob *= (overhang_set_counts[i][i] / sum(overhang_set_counts[i]))
    '''
    GG_prob = np.product(np_oh_counts.diagonal()) / \
        np.product(np_oh_counts.sum(1))

    return GG_prob


def _reduce_overhangs(breakpoints, max_num_ohs):
    """ Reduce number of overhangs in breakpoints

    The number of overhangs in each value of the breakpoints dictionary is
    reduced to less than or equal to max_num_ohs. The most orthogonal
    breakpoints are kept based on the logic of that set being the most
    versitile later on.

    Args:
        breakpoints: Dictionary of int: list pairs where keys are valid
            breakpoint indices and values are a list of valid overhangs for the
            given breakpoint represented by (overhang index, overhang sequence)
            tuples.
        max_num_ohs: int or none specifying maximum number of overhangs in each
            value of the breakpoints directory or no limit, respectively.
    """

    progress = general_tools.ProgressOutput(len(breakpoints))
    for itr, (bp, overhangs) in enumerate(breakpoints.items()):
        progress.update(itr)
        if len(overhangs) <= max_num_ohs:
            continue

        # get rid of redundant overhangs at different positions
        redundant_ohs = {}
        for oh_pos, oh_seq in overhangs:
            try:
                redundant_ohs[oh_seq].append(oh_pos)
            except KeyError:
                redundant_ohs[oh_seq] = [oh_pos]
        overhangs = [(choice(v), k) for k, v in redundant_ohs.items()]

        # find orthogonal set of overhangs
        max_p = 0
        for combo in itertools.combinations(overhangs, max_num_ohs):
            overhang_seqs = [oh_seq for _, oh_seq in combo]
            p = _calculate_GG_prob(overhang_seqs)
            if p > max_p:
                max_set = combo
                if max_p == 1.0:
                    break
                max_p = p

        if max_p:
            breakpoints[bp] = max_set
        else:
            breakpoints[bp] = sample(overhangs)


def find_GG_breakpoints(alignment, vector_overhangs, max_num_ohs=None):
    """ Finds valid Golden Gate sites to place breakpoints.

    Valid places to put breakpoints are where Golden Gate assembly can occur
    between adjacent blocks from any two parents. These sites may be considered
    as potential block endpoints for the SCHEMA-RASPP algorithm. Possible
    overhangs for each breakpoint are also returned as dictionary values.

    Args:
        alignment: List of tuples that contain the amino acid of each parent
            in the MSA at the position corresponding to the tuple index.
        vector_overhangs: Tuple of tuples containing information about the
            vector overhangs. First and second elements of outer tuple are the
            start and end overhangs, respectively. Each overhang is of the form
            (overhang index, overhang sequence).
        max_num_ohs: int or none specifying maximum number of overhangs in each
            value of the breakpoints directory or no limit, respectively.

    Returns:
        Dictionary of int: list pairs where keys are valid breakpoint indices
            and values are a list of valid overhangs for the given breakpoint
            represented by (overhang index, overhang sequence) tuples.
    """
    acceptable_overhangs = [oh for oh in all_overhangs
                            if acceptable_overhang(oh, vector_overhangs)]

    breakpoints = {0: [vector_overhangs[0]]}  # start with beginning

    for bp_num in range(1, len(alignment)):
        AA1 = set(alignment[bp_num-1])  # the last postion of block
        AA2 = set(alignment[bp_num])  # the first position of the next block

        if '-' in AA1 or '-' in AA2:  # can't recombine in a loop
            continue

        # codons that could give each AA in AA1 and AA2
        pos1_codons = [general_tools.rev_code[p] for p in AA1]
        pos2_codons = [general_tools.rev_code[p] for p in AA2]

        # overhang patterns that can give all AAs
        pos1_patterns = _get_valid_patterns(pos1_codons, reverse=True)
        pos2_patterns = _get_valid_patterns(pos2_codons)

        bp_overhangs = _patterns_to_overhangs(pos1_patterns, pos2_patterns,
                                              acceptable_overhangs)
        if not bp_overhangs:
            continue

        breakpoints[bp_num] = bp_overhangs

    breakpoints[len(alignment)] = [vector_overhangs[1]]

    if max_num_ohs:
        _reduce_overhangs(breakpoints, max_num_ohs)

    return breakpoints


def generate_weighted_E_matrix(alignment, weighted_contacts):
    """ Generate the SCHEMA energy matrix.

    The matrix is a square matrix with the same size as the length of
    alignment. The calculation is weighted by the values in the contacts dict.
    This could be the contact frequency. To get the unweighted behaviour, just
    set all contact weights=1.

    Args:
        alignment: List of tuples that contain the amino acid of each parent
            in the MSA at the position corresponding to the tuple index.
        weighted_contacts: Dictionary of tuple: float pairs where keys are
            a tuple with the indices of the two contacting residues and values
            are the weight associated with the contacts.

    Returns:
        SCHEMA energy matrix, used for calculating the SCHEMA energy of blocks.
    """
    E_matrix = np.zeros((len(alignment), len(alignment)))
    for (i, j), weight in weighted_contacts.items():
        AAs_i, AAs_j = alignment[i], alignment[j]
        parental = set(zip(AAs_i, AAs_j))  # get parental pairs of AAs at i,j
        broken = 0
        for combo in itertools.product(AAs_i, AAs_j):
            if combo not in parental:
                broken += 1  # contact is broken, record it
        E_matrix[i, j] = (broken * weight) / (len(AAs_i)**2)
    return E_matrix


def generate_blocks(breakpoints, minBL=10, maxBL=100):
    """Finds all blocks allowed by the block constraints minBL and maxBL.

    Args:
        breakpoints: Dictionary of int: list pairs from find_GG_breakpoints.
        minBL: Int minimum length of blocks.
        maxBl: Int maximum length of blocks.

    Returns:
        List of tuples that contain the start and end breakpoints of each
            block found.
    """
    blocks = set()
    for bp1, bp2 in itertools.combinations(breakpoints, r=2):
        if minBL <= bp2-bp1 <= maxBL:
            blocks.add((bp1, bp2))
    return blocks


def _block_energy(start, end, E_matrix):
    """ Calculate the SCHEMA energy of a block.

    Args:
        start: Int block's start breakpoint index.
        end: Int block's end breakpoint index.
        E_matrix: SCHEMA energy matrix from generate_weighted_E_matrix.
    """
    energy = E_matrix[start:end, end:].sum()
    return energy


# Represents node in networkx graph.
Node = collections.namedtuple('Node', ['col', 'bp'])


def _graph_column(G, blocks, E_matrix, curr_col):
    """ Add a column to the SCHEMA-RASPP graph.

    Column i represents the ith chimera breakpoint. Nodes within the column
    represent the breakpoint position within the sequence. Edges entering
    this column are the ith blocks in the chimeric sequences and are weighted
    by SCHEMA energy. Adding a column represents adding a new block to the
    possible chimeric sequences.

    Args:
        G: SCHEMA-RASPP graph.
        blocks: List of tuples from generate_blocks.
        E_matrix: SCHEMA energy graph from generate_weighted_E_matrix.
        column: Int index of column to be added.
    """
    prev_col = curr_col - 1
    previous_bps = set(node.bp for node in G if node.col == prev_col)
    for start, end in blocks:
        if start in previous_bps:
            start_node = Node(prev_col, start)
            end_node = Node(curr_col, end)
            G.add_edge(start_node, end_node, weight=_block_energy(start, end,
                                                                  E_matrix))


def _remove_deadends(G):
    """ Remove edges that do not have output edges.

    Args:
        G: SCHEMA_RASPP graph.
    """
    num_bl = max(n.col for n in G)
    for col in range(num_bl-1, 0, -1):
        col_bps = set(node.bp for node in G if node.col == col)
        for bp in col_bps:
            node = Node(col, bp)
            if G.out_degree(node) == 0:
                G.remove_node(node)


def _construct_graph(num_bl, blocks, E_matrix):
    """ Make the SCHEMA-RASPP graph.

    Represents the space of chimeric sequence as a directed graph. Graph nodes
    represent breakpoints are arranged in columns represented breakpoint
    indices in the chimeric sequences. Edges connect a given node to a node in
    the next column and represent blocks. Paths through the graph represent
    specific breakpoint positions of a resulting chimeric library. View Fig. 2
    from Endelman JB et al. "Site-directed protein recombination as a
    shortest-path problem." (2004) for a visualization.

    Args:
        num_bl: Int number of blocks in the resulting libraries.
        blocks: List of tuples from generate_blocks.
        E_matrix: SCHEMA energy graph from generate_weighted_E_matrix.

    Returns:
        G: constructed SCHEMA-RASPP graph.
        first_node: The root node at column 0 in G.
        last_node: The final node at column num_bl in G.
    """
    G = networkx.digraph.DiGraph()

    # make column 0
    first_node = Node(0, 0)
    G.add_node(first_node)

    # make middle columns
    for column in range(1, num_bl):
        _graph_column(G, blocks, E_matrix, column)

    # make last column
    column = num_bl
    last_bp = max(b[1] for b in blocks)
    ending_blocks = [b for b in blocks if b[1] == last_bp]
    _graph_column(G, ending_blocks, E_matrix, column)
    try:
        last_node = [node for node in G if node.col == column][0]
    except IndexError:
        print('Graph cannot be constructed. Try using a smaller minBL or less '
              'blocks.')
        sys.exit()

    # can remove nodes that don't have outgoing edges
    _remove_deadends(G)

    return G, first_node, last_node


def _path_energy(G, path):
    """ Calculate the energy of a path in G. Represents the SCHEMA energy of a
    library. """
    return sum(G.get_edge_data(n1, n2)['weight'] for n1, n2
               in zip(path[:-1], path[1:]))


def _remove_blocks(G, cutoff_len, min_max):
    """ Remove blocks from the graph based on a cutoff length.

    Args:
        G: SCHEMA-RASPP graph.
        cutoff_len: Int block length at which exceeding blocks are removed.
        min_max: 'max' or 'min' stating whether cutoff_len is a maximum or
            minimum cutoff.
    """
    removal_edges = []
    for edge in G.edges:
        node1, node2 = edge
        block_length = node2.bp - node1.bp
        if min_max == 'min' and block_length < cutoff_len:
            removal_edges.append(edge)
        elif min_max == 'max' and block_length > cutoff_len:
            removal_edges.append(edge)
    G.remove_edges_from(removal_edges)

    # can remove nodes that don't have outgoing edges
    _remove_deadends(G)

    # remove nodes with no edges
    removal_nodes = [n for n in G if G.degree(n) == 0]
    G.remove_nodes_from(removal_nodes)


def _find_shortest_path(G, first_node, last_node, libraries):
    """ Find the shortest path through G from first_node to last_node and
    update the SCHEMA energy of the libraries with the culmulative edge energy
    through the path. """
    path = networkx.dijkstra_path(G, first_node, last_node)
    energy = _path_energy(G, path)
    bps = tuple(node.bp for node in path)
    libraries[bps] = {'energy': energy}


def shortest_path_recombination(num_bl, blocks, E_matrix):
    """ Find libraries with minimum energy with block length constraints.

    Scan over all pairs of minimum and maximum block lengths, using a
    SCHEMA-RASPP graph and Dijkstra's algorithm to find the minimum energy
    library for each pair.

    Args:
        num_bl: Int number of blocks in the resulting libraries.
        blocks: List of tuples from generate_blocks.
        E_matrix: SCHEMA energy graph from generate_weighted_E_matrix.

    Returns:
        Dictionary of lib_bps: {'E':' <SCHEMA energy>} pairs where lib_bps is a
            tuple of breakpoint indices for a library.
    """

    masterG, first_node, last_node = _construct_graph(num_bl, blocks, E_matrix)

    align_len = max([end for _, end in blocks])
    minBL = min(end - start for start, end in blocks)
    maxBL = max(end - start for start, end in blocks)
    libraries = {}

    # solve shortest path using dijkstra algorithm
    _find_shortest_path(masterG, first_node, last_node, libraries)

    # scan over all minBL+maxBL combinations
    # scan over min block sizes from minBL to seq_len/num_blocks
    equal_size_blocks = align_len // num_bl
    print(f'Total iterations: {equal_size_blocks-minBL} (later iterations'
          ' are faster)', flush=True)
    for min_cutoff in range(minBL, equal_size_blocks + 1):
        print(f'Current iteration: {min_cutoff-minBL}\r', end='', flush=True)
        G = masterG.copy()
        _remove_blocks(G, min_cutoff, min_max='min')
        # scan over max block sizes from maxBL till it can't solve the SPP
        for max_cutoff in range(maxBL, 0, -1):
            _remove_blocks(G, max_cutoff, min_max='max')
            try:
                _find_shortest_path(G, first_node, last_node, libraries)
            except (networkx.exception.NetworkXNoPath,
                    networkx.exception.NodeNotFound):
                # this happens once there are no paths
                break

    print(flush=True)
    return libraries


def _get_block_alignment(lib_bps, alignment):
    """ Assign block numbers to each element in alignment.

    Args:
        lib_bps: Tuple of ints represent library breakpoint indices. Keys from
            libraries.
        alignment: List of tuples that contain the amino acid of each parent
            in the MSA at the position corresponding to the tuple index.

    Returns:
        List of tuples. Same as alignment but each tuple has the block number
            of the associated residue at position 0.
    """
    block_alignment = []
    start_bps, end_bps = lib_bps[:-1], lib_bps[1:]
    for blnum, (start, end) in enumerate(zip(start_bps, end_bps)):
        extension = [(blnum,) + AAs for AAs in alignment[start:end]]
        block_alignment.extend(extension)
    return block_alignment


def _get_binned_blocks(block_alignment):
    """ Bins parental block sequences by position.

    Args:
        block_alignment: List of tuples returned by _get_block_alignment.

    Returns:
        List of lists of block sequences in bins. Bins are represented by each
            inner list, which are indexed according to the block location of
            block sequences they hold.
    """
    block_bins = {}
    for blk_num, *AAs in block_alignment:
        if blk_num in block_bins:
            block_bins[blk_num].append(AAs)
        else:
            block_bins[blk_num] = [AAs]
    block_bins = [zip(*blk_bin) for blk_bin in block_bins.values()]
    binned_blocks = [[''.join(pb) for pb in blk_bin] for blk_bin in block_bins]
    return binned_blocks


def _chimera_muts(chimera_seq, parent_seq):
    """ Calculates the number of mutations required to convert chimera_seq
    to parent_seq. """
    return sum(1 for cAA, pAA in zip(chimera_seq, parent_seq) if cAA != pAA)


def _calculate_M(block_alignment, sample=False):
    """ Calculate the average mutational diversity in a library.

    Args:
        block_alignment: List of tuples returned by _get_block_alignment.

    Return:
        float average number of mutations from nearest parent in library.
    """
    block_seq, *parent_lists = zip(*block_alignment)
    parents = [''.join(par) for par in parent_lists]
    block_nums = sorted(set(block_seq))
    binned_blocks = _get_binned_blocks(block_alignment)

    totalM = 0
    for ch in itertools.product(*binned_blocks):
        chimera_seq = ''.join(ch)
        totalM += min(_chimera_muts(chimera_seq, ps) for ps in parents)
    M = totalM / len(parents) ** len(block_nums)
    return M


def update_M(libraries, alignment):
    """ Update libraries with average number of mutations from nearest parent.

    Args:
        libraries: Dictionary of lib_bps: {'E':' <SCHEMA energy>} pairs where
            lib_bps is a tuple of breakpoint indices for a library. Created
            by shortest_path_recombination.
        alignment: List of tuples that contain the amino acid of each parent
            in the MSA at the position corresponding to the tuple index.
    """
    progress = general_tools.ProgressOutput(len(libraries))
    for itr, (lib_bp, lib_attrs) in enumerate(libraries.items()):
        progress.update(itr)
        block_alignment = _get_block_alignment(lib_bp, alignment)
        M = _calculate_M(block_alignment)
        lib_attrs['M'] = M


def update_GG_prob(libraries, breakpoints):
    """ Updates libraries with Golden Gate information.

    Each library is updated with the set of overhangs with highest GG_prob and
    that GG_prob. Libraries with low GG_prob are likely to have issues in
    Golden Gate reactions.

    Args:
        libraries: Dictionary of lib_bps: {'E':' <SCHEMA energy>} pairs where
            lib_bps is a tuple of breakpoint indices for a library. Created
            by shortest_path_recombination.
        breakpoints: Dictionary of int: list pairs from find_GG_breakpoints.
    """
    progress = general_tools.ProgressOutput(len(libraries))
    for itr, (bp_positions, lib_attrs) in enumerate(libraries.items()):
        progress.update(itr)
        lib_breakpoints = {bp_pos: breakpoints[bp_pos] for bp_pos
                           in bp_positions}
        max_gg_prob = 0
        for oh_set in itertools.product(*lib_breakpoints.values()):
            overhangs = [pos_oh[1] for pos_oh in oh_set]
            gg_prob = _calculate_GG_prob(overhangs)
            if gg_prob > max_gg_prob:  # better overhang set found
                max_gg_prob = gg_prob
                max_set = oh_set
                if gg_prob == 1:
                    break

        lib_attrs.update({'GG_prob': max_gg_prob})
        if lib_attrs['GG_prob'] != 0:
            lib_attrs['GG_sites'] = max_set

    return libraries



import random
from subprocess import Popen
import sys
import urllib

from Bio import Seq, SeqRecord, SeqIO
from Bio.SeqUtils import seq1

# List of optimal codons for E. coli.
AA_C31 = {'A': ('GCT', 'GCA'), 'R': ('CGT', 'CGA'), 'N': ('AAT',),
          'D': ('GAT',), 'C': ('TGT',), 'Q': ('CAA', 'CAG'), 'E': ('GAA',),
          'G': ('GGT',), 'H': ('CAT', 'CAC'), 'I': ('ATT', 'ATC'),
          'L': ('TTA', 'TTG', 'CTA'), 'K': ('AAA',), 'M': ('ATG',),
          'F': ('TTT',), 'P': ('CCT', 'CCA'), 'S': ('AGT', 'TCA'),
          'T': ('ACA', 'ACT'), 'W': ('TGG',), 'Y': ('TAT',),
          'V': ('GTT', 'GTA'), '*': ('TGA',), '-': ('---',)}


def get_pdb(sequence):
    """ Find and download the closestPDB sequence to sequence.

    Uses a BLAST search of the PDB database to find the PDB sequence closest to
    the input sequence. The BLAST search sequenceIdentityCutoff is scanned from
    0 to 100 with a binary search algorithm until a unique pdb is found. If
    no unique PDB is found, but one exists, the returned PDBs all have the same
    percent identity with sequence and any can be chosen.

    Documentation for PDB API: https://www.rcsb.org/pages/webservices. It
    would have been more efficient to use the direct BLAST-PDB API
    (https://www.rcsb.org/pdb/software/rest.do#blastPdb) to return BLAST search
    sequence identities, but it did not return the best sequences for a bgl3
    example. I may have been doing something wrong or that API is broken.

    Args:
        sequence: Query sequence for BLAST search

    Returns:
        pdb filename of (possibly one of) the database sequence with highest
            pairwise identity to sequence, where pdb filename was downloaded
    """

    url = 'http://www.rcsb.org/pdb/rest/search'
    lower_bound, upper_bound = 0, 100  # adjust cutoff with average of these
    pdb_identities = [False for _ in range(101)]  # list of pdbs at each iden %
    while True:
        identity_cutoff = (upper_bound + lower_bound) // 2
        print("Current identity percentage: %s" % identity_cutoff, flush=True)

        # pdb-blast search parameters
        queryText = """
        <orgPdbQuery>
            <queryType>org.pdb.query.simple.SequenceQuery</queryType>
            <sequence>%s</sequence>
            <searchTool>blast</searchTool>
            <maskLowComplexity>yes</maskLowComplexity>
            <sequenceIdentityCutoff>%s</sequenceIdentityCutoff>
        </orgPdbQuery>""" % (sequence, identity_cutoff)

        # make request and read output
        req = urllib.request.Request(url, data=queryText.encode())
        f = urllib.request.urlopen(req)
        result = f.read().decode()
        pdbs = result.strip().split('\n')
        pdb_identities[identity_cutoff] = pdbs

        # if unique pdb found, select it and break
        if (len(pdbs) == 1 and pdbs != ['']):
            pdb = pdbs[0]
            break

        # need another refinement round, if no pdbs, avg is too high. if more
        # than 1 pdb, avg is too low, save a pdb for possible break next round
        if pdbs == ['']:
            upper_bound = identity_cutoff
        elif len(pdbs) > 1:
            lower_bound = identity_cutoff
            pdb = pdbs[0]

        # identity % already tested, search is done. pdb was set to first of
        # pdbs for largest identity % when lower_bound was last reset
        next_cutoff = (upper_bound + lower_bound) // 2
        if pdb_identities[next_cutoff]:
            break

    if not pdb:
        print('no pdb found')
        sys.exit()
    else:
        print(f'pdb found: {pdb}')

    # pdb is format "1GNX:1", not sure what the ":1" is but remove it
    pdb = pdb.split(':')[0]

    # retrieve the pdb file
    pdb_fn = pdb + '.pdb'
    url = 'http://www.rcsb.org/pdb/files/' + pdb_fn
    with urllib.request.urlopen(url) as response, \
            open(pdb_fn, 'wb') as out_file:
        data = response.read()
        out_file.write(data)

    return pdb_fn


class atom(object):
    def __init__(self, line):
        self.line = line
        self.resName = line[17:20].strip()
        self.resSeq = int(line[22:26].strip())
        self.x = float(line[30:38].strip())
        self.y = float(line[38:46].strip())
        self.z = float(line[46:54].strip())

    def __str__(self):
        line = list(self.line)
        line[22:26] = list('{:4d}'.format(self.resSeq))
        return ''.join(line)


def d(a1, a2):
    return ((a1.x - a2.x)**2 + (a1.y - a2.y)**2 + (a1.z - a2.z)**2) ** (1/2)


class residue(object):
    def __init__(self, atom):
        self.atoms = [atom]
        self.resName = atom.resName
        self._resSeq = atom.resSeq

    def add_atom(self, atom):
        self.atoms.append(atom)

    def d(self, res):
        m = d(self.atoms[0], res.atoms[0])
        for i in self.atoms:
            for j in res.atoms:
                dis = d(i, j)
                if dis < m:
                    m = dis
        return m

    @property
    def resSeq(self):
        return self._resSeq

    @resSeq.setter
    def resSeq(self, new_resSeq):
        for a in self.atoms:
            a.resSeq = new_resSeq
        self._resSeq = new_resSeq

    def __repr__(self):
        return f'residue: {self.resName} {self.resSeq}'


def run_muscle(in_fn, out_fn):
    # Align sequences with muscle
    print(f'Running muscle with {in_fn}, saving to {out_fn}.', flush=True)
    Popen(f'muscle -in {in_fn} -out {out_fn}',
          shell=True).wait()
    out_SRs = {SR.name: SR for SR in SeqIO.parse(out_fn, 'fasta')}
    in_labels = [SR.name for SR in SeqIO.parse(in_fn, 'fasta')]
    stable_out_SRs = [out_SRs[label] for label in in_labels]
    SeqIO.write(stable_out_SRs, out_fn, 'fasta')
    print('\n\nMuscle done.')


def read_PDB(pdb_fn):
    with open(pdb_fn) as f:
        atoms = [atom(i.strip()) for i in f if 'ATOM' in i[:4]]
    AAs = {}
    for a in atoms:
        if a.resSeq not in AAs:
            AAs.update({a.resSeq: residue(a)})
        else:
            AAs[a.resSeq].add_atom(a)
    return AAs.values()


def write_PDB(residues, fn):
    with open(fn, 'w') as fn:
        for AA in residues:
            if AA.resSeq is not None:
                for atom in AA.atoms:
                    fn.write(str(atom) + '\n')


def get_PDB_seq(AAs):
    return ''.join([seq1(AA.resName) for AA in AAs])


def save_SeqRecords(seqs, names, handle):
    seqs = [Seq.Seq(sequence) for sequence in seqs]
    zipped_SRs = zip(seqs, names)
    records = [SeqRecord.SeqRecord(s, id=n) for s, n in zipped_SRs]
    SeqIO.write(records, handle, 'fasta')


def get_pos_CDNs(pos):
    single_AAs = set(pos)
    translation = {AA: random.choice(AA_C31[AA]) for AA in single_AAs}
    return tuple(translation[AA] for AA in pos)



""" Align a group of sequences and calculate a contact map.

Run syntax: "python step1.py <seq_fn> <pdb_fn> <out_prefix>"
Example: "python step1.py bgl3_parents.fasta 1gnx.pdb bgl3"

Command Line Args:
    seq_fn: name of sequences file read by muscle
    pdb_fn: name of PDB file used to make contact map. If "search", script will
        search PDB database and find structure with closest identity to first
        parent sequence
    out_prefix: prefix on output files

Outputs:
    <out_prefix>_AA.fasta: name of aligned amino acid sequences file, used in
        step 2 and step 4
    <out_prefix>_contacts.json: name of calculated contacts file, used in
        step 2
    <out_prefix>_CDN.fasta: name of aligned codon sequences file, used in
        step 4
    renumbered_<pdb_fn>: name of renumber pdb file
    <out_prefix_PDB.fasta>: name of pdb alignment file

It takes as input a sequence fasta file and a PDB structure file. The sequence
fasta should list the parent amino acid sequences for SCHEMA-RASPP, with
(preferably only) the name of each sequence in the headers. The PDB structure
will be aligned to the output multiple sequence alignment and renumbered
accordingly.

The main output files are "<out_prefix>_AA.fasta",
"<out_prefix>_contacts.json", and "<out_prefix>_CDN.fasta", which are used in
later steps. "renumbered_<pdb_fn>" and "<out_prefix>_PDB.fasta" are also output
for logging purposes.

You can also replace all instances of "sys.argv" in the code with the input
filenames directly, then run "python step1.py".

Contacts are defined as pairs of residues in the PDB file that have a distance
of 4.5 angstroms measured at their closest atoms. <pdb_fn> must be in PDB
format, but is not required to be from the Protein Data Bank. Programs such
as MODELLER can generate predicted structures for a given sequence, which
can then be used as an input pdb file.

BioPython and Muscle are required packages. The Romero Lab group server has
both installed.
"""

import json
import os
import sys

from Bio import Seq, SeqRecord, SeqIO

from tools import step1_tools

seq_fn = sys.argv[1]  # name of sequences file read by muscle
pdb_fn = sys.argv[2]  # name of PDB file used to make contact map
out_prefix = sys.argv[3]  # prefix of output files

if pdb_fn == 'search':
    print('Searching pdb database for closest structure match to first parent',
          flush=True)
    p1 = list(SeqIO.parse(seq_fn, 'fasta'))[0]
    pdb_fn = step1_tools.get_pdb(str(p1.seq))

step1_tools.run_muscle(seq_fn, out_prefix + '_AA.fasta')

# Load muscle output fasta file
print('Loading muscle output', flush=True)
alignment_SRs = list(SeqIO.parse(f'{out_prefix}_AA.fasta', 'fasta'))
names = [str(i.id) for i in alignment_SRs]
AA_seqs = [str(i.seq) for i in alignment_SRs]
AA_alignment = list(zip(*AA_seqs))  # list of AAs at each position

# Load sequence from pdb and temporarily add to AA alignment
print('Loading pdb file', flush=True)
pdb_AAs = step1_tools.read_PDB(pdb_fn)
pdb_Seq = step1_tools.get_PDB_seq(pdb_AAs)
pdb_Seq = Seq.Seq(pdb_Seq)
pdb_SeqRecord = SeqRecord.SeqRecord(pdb_Seq, name='PDB')
temp_SRs = [pdb_SeqRecord] + alignment_SRs

# Align pdb sequence to AA aligned sequences
print('Aligning pdb sequence to AA aligned sequencing', flush=True)
SeqIO.write(temp_SRs, 'temp_seqs.fasta', 'fasta')
step1_tools.run_muscle('temp_seqs.fasta', f'{out_prefix}_PDB.fasta')
os.remove('temp_seqs.fasta')
pdb_align_SRs = SeqIO.parse(f'{out_prefix}_PDB.fasta', 'fasta')
pdb_align_seqs = [str(i.seq) for i in pdb_align_SRs]
pdb_alignment = list(zip(*pdb_align_seqs))

# Renumber pdb residues and save new pdb
print('Renumbering pdb file', flush=True)
pdb_AAs_residue = iter(pdb_AAs)
pos = 0    # PDB will be 1-indexed for consistency with other PDB files
for pdb_AA, *parent_AAs in pdb_alignment:
    valid_parent_pos = any([a != '-' for a in parent_AAs])
    if valid_parent_pos:
        pos += 1
    if pdb_AA != '-':
        AA = next(pdb_AAs_residue)
        assert step1_tools.seq1(AA.resName) == pdb_AA
        if valid_parent_pos:
            AA.resSeq = pos
        else:
            AA.resSeq = None
step1_tools.write_PDB(pdb_AAs, 'renumbered_' + pdb_fn)

# Calculate contacts
print('Calculating contacts', flush=True)
pdb_AAs = list(step1_tools.read_PDB('renumbered_' + pdb_fn))
for AA in pdb_AAs:
    AA.resSeq -= 1   # make index start from 0
contacts = []
for index, AA1 in enumerate(pdb_AAs):
    for AA2 in pdb_AAs[index+1:]:
        if AA1.d(AA2) < 4.5:
            contacts.append((AA1.resSeq, AA2.resSeq))
with open(out_prefix + '_contacts.json', 'w') as f:
    json.dump(contacts, f)

# Add stop codons if necessary
if AA_alignment[-1] != tuple(['*'] * len(alignment_SRs)):
    AA_alignment.append(tuple(['*'] * len(alignment_SRs)))

# Make and save codon alignments
print('Making codon alignment', flush=True)
CDN_align = [step1_tools.get_pos_CDNs(pos) for pos in AA_alignment]
CDN_seqs = [''.join(CDN_pos) for CDN_pos in zip(*CDN_align)]
step1_tools.save_SeqRecords(CDN_seqs, names, f'{out_prefix}_CDN.fasta')



""" Choose a library based on SCHEMA energy and mutational levels.

Run syntax: "python step3.py <libraries_fn> <chosen_lib_fn>"
Example: "python step3.py bgl3_libraries.json bgl3_chosen_lib.json"

Command Line Args:
    libraries_fn: name of libraries file generated in step 2
    gg_prob_thresh: lower bound of gg_prob allowed for accepted libraries
    chosen_lib_fn: name of output chosen library file

Outputs:
    chosen library in <chosen_lib_fn>, used in step 4
    libraries.png: graph of SCHEMA-RASPP curve with chosen_library


You can also replace all instances of "sys.argv" in the code with the input
filenames directly, then run "python step2.py".

As written, the script automatically chooses the library with maximum m-e where
m is the average number of mutations and e is the SCHEMA energy for the
library. To manually select a library, comment out the automatic selection,
find the breakpoints key for the desired library, then save the desired library
key-value pair in a tuple for json packaging.

Matplotlib is a required package. The Romero lab group server has it installed.
"""

import json
import sys

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

libraries_fn = sys.argv[1]
gg_prob_threshold = float(sys.argv[2])
chosen_lib_fn = sys.argv[3]

with open(libraries_fn, 'r') as f:
    libraries = {tuple(k): v for k, v in json.load(f)}


gg_filtered_libs = {k: v for k, v in libraries.items() if v['GG_prob'] >= 0.95}

if not gg_filtered_libs:
    print('No acceptable Golden Gate libraries found with a GG_prob threshold '
          f'of {gg_prob_threshold}')
    sys.exit()

chosen_lib_bps = max(gg_filtered_libs, key=lambda x: gg_filtered_libs[x]['M']
                     - gg_filtered_libs[x]['energy'])
chosen_lib_attrs = libraries[chosen_lib_bps]

graph_data = [(l['M'], l['energy'], l['GG_prob']) for l in libraries.values()]
graph_chosen_lib = (chosen_lib_attrs['M'], chosen_lib_attrs['energy'])

plt.ioff()
plt.figure(figsize=(16, 9))
m, e, g = zip(*graph_data)
plt.scatter(m, e, c=g, vmax=1, vmin=.95)
plt.annotate('chosen library', graph_chosen_lib)
plt.colorbar().set_label('GG_prob', rotation=270, labelpad=12)
plt.title('Libraries')
plt.xlabel('M')
plt.ylabel('energy')
plt.savefig('libraries.png')

chosen_lib = (chosen_lib_bps, chosen_lib_attrs)
with open(chosen_lib_fn, 'w') as f:
    json.dump(chosen_lib, f)



"""Module for running the RASPP algorithm.

TODO: Refactor to separate codon_options, EnergyFunctions.
TODO: Add ABC for problem (RASPP superclass). Zheng et al. 2009 has joint
optimization of stability and diversity. Thus there are more algorithms that
may be useful to implement/replace RASPP. Does this necessitate renaming the
package to something like ggrecomb?
"""

from collections.abc import Iterable
from dataclasses import dataclass, field
from itertools import product, combinations
from typing import NamedTuple, Optional, Type, Union

from ggrecomb.energy_functions import EnergyFunction, SCHEMA
from ggrecomb.library import RecombinantLibrary
from ggrecomb import ParentAlignment
from ggrecomb import PDBStructure


'''
# List of optimal codons for E. coli.
AA_C31 = {'A': ['GCT', 'GCA'], 'R': ['CGT', 'CGA'], 'N': ['AAT'], 'D': ['GAT'],
          'C': ['TGT'], 'Q': ['CAA', 'CAG'], 'E': ['GAA'], 'G': ['GGT'],
          'H': ['CAT', 'CAC'], 'I': ['ATT', 'ATC'], 'L': ['TTA', 'TTG', 'CTA'],
          'K': ['AAA'], 'M': ['ATG'], 'F': ['TTT'], 'P': ['CCT', 'CCA'],
          'S': ['AGT', 'TCA'], 'T': ['ACA', 'ACT'], 'W': ['TGG'], 'Y': ['TAT'],
          'V': ['GTT', 'GTA'], '*': ['TGA'], '-': ['---']}
'''

# List of optimal codons for E. coli.
AA_C31 = {'A': ('GCT', 'GCA'), 'R': ('CGT', 'CGA'), 'N': ('AAT',),
          'D': ('GAT',), 'C': ('TGT',), 'Q': ('CAA', 'CAG'), 'E': ('GAA',),
          'G': ('GGT',), 'H': ('CAT', 'CAC'), 'I': ('ATT', 'ATC'),
          'L': ('TTA', 'TTG', 'CTA'), 'K': ('AAA',), 'M': ('ATG',),
          'F': ('TTT',), 'P': ('CCT', 'CCA'), 'S': ('AGT', 'TCA'),
          'T': ('ACA', 'ACT'), 'W': ('TGG',), 'Y': ('TAT',),
          'V': ('GTT', 'GTA'), '*': ('TGA',), '-': ('---',)}


class RASPP:
    """Recombination as a shortest path problem.

    Construct the graph depicted in Endelman et al 2004 figure 2. A node at
    index i in column c represents the selection of the (c-1)th breakpoint at
    the ith position in the multiple sequence alignment. Specifically, the ith
    amino acid in an aligned chimera will be the first residue of block c,
    while the (i-1)th amino acid is the last residue of block c-1. An edge
    between two nodes represents the block between the two breakpoints. A
    recombinant sequence library is therefore a traversal through the graph
    from node (c=0, i=0) to (n, i_f), where the sum of the edge weights in the
    path is the average energy of the library.

    Public attributes:
        columns: List of Lists of Nodes representing graph.
        breakpoints: Available recombination cross-over locations.

    Public methods:
        eval_bps: Average SCHEMA energy of sequence library with given
            breakpoints.
        eval_all_bps: eval_bps over all combinations of bps.
        min_bps: Find the set of breakpoints with minimum energy.
    """
    def __init__(
            self,
            parent_aln: ParentAlignment,
            n: int,
            start_overhang: Optional[tuple[int, str]] = None,
            end_overhang: Optional[tuple[int, str]] = None,
            codon_options: dict[str, tuple[str, ...]] = AA_C31,
            energy_func: Union[EnergyFunction, Type[EnergyFunction]] = SCHEMA
    ):
        """Init a RASPP graph.

        Args:
            parent_aln: Parent multiple sequence alignment.
            n: Number of breakpoints (crossovers) in output libraries.
            start_overhang: Positional shift and nucleotide sequence of Golden
                Gate site for vector insertion at start of sequence. Not
                factored into calculations if None.
            end_overhang: Positional shift and nucleotide sequence of Golden
                Gate site for vector insertion at end of sequence. Not factored
                into calculations if None.
            codon_options: Amino acid to available codon mapping. Used in
                Golden Gate site design and library sequence design. Change
                this to include or exclude certain codons based on codon
                optimization schemes, reassigned codons, etc.
            energy_func: Energy function used in calculations. See
                EnergyFunction for more information. May pass in class or
                instance.
        """

        # Calculate all crossover sites valid for Golden Gate Assembly.
        self.breakpoints, self._bp_to_group = _calculate_breakpoints(
            parent_aln, codon_options, start_overhang, end_overhang)

        if isinstance(energy_func, type):
            energy_func = energy_func(parent_aln)

        # Cache used for redundant <M> calculations. Certain adjacent
        # breakpoints will have the same <M>, e.g. if there is only one amino
        # acid in the alignment at each of the two breakpoint positions. Use
        # the self._bp_to_group map to convert a library's breakpoints to the
        # nonredundant breakpoint group indices, which will be the keys to
        # self._bp_group_M_cache.
        self._bp_group_M_cache: dict[tuple[int, ...], float] = {}

        self.pa = parent_aln

        # Build RASPP graph nodes.
        if start_overhang is None:
            start_overhang_list = None
        else:
            start_overhang_list = [start_overhang]
        first_col = [_Node(0, 0, start_overhang_list)]
        self.columns = [first_col]
        for col in range(1, n + 1):
            # No edges into node => exclude it.
            min_index = self.columns[col-1][0].index + 1
            new_col = [_Node(col, index, self.breakpoints[index])
                       for index in self.breakpoints if index >= min_index]
            self.columns.append(new_col)

        # Fill out edges between nodes.
        self.columns[0][0].fill_out(self.columns[1], energy_func)
        for col, next_col in zip(self.columns[1:], self.columns[2:]):
            for node in col[:-1]:
                node.fill_out(next_col, energy_func)

    # TODO: Refactor or remove next two methods to be compatible?
    def eval_bps(self, bps: Iterable[int]) -> float:
        """Average SCHEMA energy of sequence library with given breakpoints."""
        e = 0.0
        curr_node = self.columns[0][0]
        for bp in bps:
            edge = curr_node.out_edges[bp]
            e += edge.e_diff
            curr_node = edge.out_node
        return e

    def eval_all_bps(self) -> dict[tuple[int], float]:
        """eval_bps over all combinations of breakpoints."""
        valid_bps = list(self.breakpoints.keys())
        try:
            del valid_bps[0]
            del valid_bps[-1]
        except KeyError:
            pass
        bps_e = {bps: self.eval_bps(bps) for bps
                 in combinations(valid_bps, len(self.columns)-1)}
        return bps_e

    def min_bps(
        self,
        min_block_len: Optional[int] = None,
        max_block_len: Optional[int] = None
    ) -> 'RecombinantLibrary':
        """Find the set of breakpoints with minimum energy."""

        n = len(self.columns) - 1  # number of breakpoints
        N = len(str(self.pa.aligned_sequences[0].seq))

        if min_block_len is None:
            min_block_len = 0
        if max_block_len is None:
            max_block_len = N

        # Memoized traversal over columns of graph to find minimum library.
        # Record the minimum library for each node during traversal.
        first_node = self.columns[0][0]
        first_lib = _TempLibrary(0.0, first_node.breakpoint)
        next_col_min_libs = {first_node: first_lib}
        while next_col_min_libs:
            curr_col_min_libs = next_col_min_libs

            # print(list(curr_col_min_libs)[0].col)
            if list(curr_col_min_libs)[0].col == n:
                # We've reached the last column.
                break

            # Mapping from each node in the next column to it's minimum energy
            # library.
            next_col_min_libs = {}

            # Iterate over every node and it's minimum energy library in the
            # current column to find the minimum energy library for each node
            # in the next column.
            for curr_node, curr_min_lib in curr_col_min_libs.items():
                # Iterate over all edges from curr_node. Each edge results in
                # a candidate library built from curr_lib and the breakpoint
                # from the node at the end of the edge.
                for edge in curr_node.out_edges:
                    new_node = edge.out_node

                    # Check that the new block is of valid size.
                    next_block_len = new_node.index - curr_node.index
                    if not min_block_len <= next_block_len <= max_block_len:
                        continue

                    # Get the best library for new_node so far and compare to
                    # library from curr_node. If no best library or curr_node
                    # library is best, set as the best library.
                    best_lib = next_col_min_libs.get(new_node)
                    if best_lib is None or curr_min_lib.energy + edge.e_diff \
                       < best_lib.energy:
                        new_lib = curr_min_lib.expand(edge.e_diff,
                                                      new_node.breakpoint)
                        next_col_min_libs[new_node] = new_lib

        candidate_libs = [lib for node, lib in curr_col_min_libs.items()
                          if min_block_len <= N - node.index <= max_block_len]
        if not candidate_libs:
            raise LibrariesNotFoundError
        return_lib = min(candidate_libs, key=lambda lib: lib.energy)
        if not return_lib.breakpoints:
            raise LibrariesNotFoundError

        # add last breakpoint
        last_bp = {N: self.breakpoints[N]}
        temp_lib = return_lib.expand(0.0, last_bp)
        e = temp_lib.energy
        bps = temp_lib.breakpoints
        return_lib = RecombinantLibrary(e, bps, self.pa, self._bp_to_group,
                                        self._bp_group_M_cache)
        return return_lib

    def vary_m_proxy(self, L_min, L_max):
        """Choose E-optimal libraries by varying block length as an M proxy.

        TODO: Erase/refactor to get rid of old/naive min_bps() and
        calc_average_m().
        """
        chosen_libs = []

        # TODO: Can probably do problem reduction for each min_block_len
        # iteration. E.g. For a given min_block_len iteration, iterating
        # through max_block_len backwards allows us to say the most recently
        # found library is the minimum E until its max_block_len is greater
        # max_block_len.
        # TODO: DOUBLE CHECK THIS
        for min_block_len in range(L_min, L_max):
            curr_best = None
            for max_block_len in range(L_max, min_block_len, -1):
                print(min_block_len, max_block_len, '\r', end='')
                if curr_best is not None:
                    if curr_best.max_block_len <= max_block_len:
                        continue
                    else:
                        curr_best = None
                try:
                    lib = self.min_bps(min_block_len, max_block_len)
                    curr_best = lib

                    if lib not in chosen_libs:
                        lib.calc_average_m(self.pa, self._bp_group_map,
                                           self._bp_group_M_cache)
                        chosen_libs.append(lib)
                except LibrariesNotFoundError:
                    pass

        return chosen_libs

    def all_libraries(
        self,
        min_block_len: Optional[int] = None,
        max_block_len: Optional[int] = None
    ) -> list['RecombinantLibrary']:
        """Get all possible libraries with depth-first graph traversal."""
        libs = []
        n = len(self.columns) - 1  # number of breakpoints
        N = sorted(self.breakpoints)[-1]  # length of sequences

        if min_block_len is None:
            min_block_len = 0
        if max_block_len is None:
            max_block_len = N

        first_node = self.columns[0][0]
        first_lib = RecombinantLibrary(0.0, first_node.breakpoint)
        stack = [(first_node, first_lib)]
        while stack:
            curr_node, curr_lib = stack.pop()
            if curr_node.col == n:
                next_block_len = N - curr_node.index
                if min_block_len <= next_block_len <= max_block_len:
                    last_bp = {N: self.breakpoints[N]}
                    curr_lib = curr_lib.expand(0.0, last_bp)
                    curr_lib.calc_average_m(self.pa, self.bp_group_map)
                    libs.append(curr_lib)
                continue
            for edge in reversed(curr_node.out_edges.values()):
                next_block_len = edge.out_node.index - curr_node.index
                if min_block_len <= next_block_len <= max_block_len:
                    new_node = edge.out_node
                    new_lib = curr_lib.expand(edge.e_diff, new_node.breakpoint)
                    stack.append((new_node, new_lib))
        if not libs:
            msg = f'No libraries found with min_block_len={min_block_len}' \
                  f' and max_block_len={max_block_len}.'
            raise LibrariesNotFoundError(msg)
        return libs


@dataclass
class _TempLibrary:
    energy: float
    breakpoints: dict[int, list[tuple[int, str]]] = field(default_factory=dict)

    def expand(self, e_diff: float, new_bp: dict[int, list[tuple[int, str]]]):
        new_e = self.energy + e_diff
        new_bps = self.breakpoints | new_bp
        return type(self)(new_e, new_bps)


class LibrariesNotFoundError(Exception):
    pass


def _get_valid_patterns(
    cdn_sets: tuple[set[str]],
    reverse: bool = False
) -> tuple[set[str]]:
    """DNA patterns in the codon alignment for Golden Gate site construction.

    Helper function for _calculate_breakpoints. Each input set corresponds to
    the valid codons that encode a given amino acid at a given position in the
    parent alignment.  For example, if a certain position has isoleucine and
    threonine, and the valid codons for these amino acids are ('ATT', 'ATC')
    and ('ACC', 'ACT'), respectively, ({'ATT', 'ATC'), {'ACC', 'ACT'}) will be
    passed in. The length 1, 2, and 3 patterns are returned. See the example
    below for more information.

    Args:
        cnd_sets: Collections of codons that code for each amino acids at a
            given site in the parent alignment.
        reverse: If the patterns should be explored right to left instead of
            right to left (default).

    Returns:
        Collections of all length 1, 2, and 3 patterns found.

    Example:
        _get_valid_patterns(({'ATT', 'ATC'}, {'ACC', 'ACT'})) will return
            ({'A'}, {}, {}) since 'A' can be found as the first letter at least
            once in each set. There are no length 2 patterns because there are
            no common letters at the second codon position.
        _get_valid_patterns(({'ATT', 'ATC'}, {'ACC', 'ACT'}), True) will return
            ({'C', 'T'}, {}, {}) since 'C' and 'T' can both be found as the
            last letter at least once in each set.

    TODO: Eliminate crossovers based on invalid/bad GG sites.
    """

    # If your codons are not three bases long, you're either wrong or doing
    # xenobiology.
    assert all(all(len(cdn) == 3 for cdn in cdn_set) for cdn_set in cdn_sets)

    # If not reverse, cdn_shrink removes the right side letter of each codon.
    # If reverse, cdn_shrink removes the left side letter of each codon.
    fwd_lim, bwd_lim = int(reverse), 1 - int(reverse)

    def cdn_shrink(cdns):
        return {cdn[fwd_lim:len(cdn)-bwd_lim] for cdn in cdns}

    # 1 AA per codon, so len 3 patterns exist iff len(cdn_sets) == 1.
    if len(cdn_sets) == 1:
        len_3_sets = tuple(cdn_sets)[0]
    else:
        len_3_sets = {}

    # Remove a codon letter and take intersection to get all len 2 patterns.
    cdn_sets = [cdn_shrink(cdns) for cdns in cdn_sets]
    len_2_sets = set.intersection(*cdn_sets)

    # Remove another letter and take intersection to get all len 1 patterns.
    cdn_sets = [cdn_shrink(cdns) for cdns in cdn_sets]
    len_1_sets = set.intersection(*cdn_sets)

    return len_1_sets, len_2_sets, len_3_sets


def _calculate_breakpoints(
    parent_aln: ParentAlignment,
    codon_options: dict[str, tuple[str, ...]],
    start_overhang: Optional[tuple[int, str]] = None,
    end_overhang: Optional[tuple[int, str]] = None,
) -> dict[int, list[tuple[int, str]]]:
    """Calculate the breakpoints for the parent alignment.

    Currently supports length four Golden Gate sites.

    # TODO: move this to module header?
    A breakpoint is potential site for recombination of the parental sequences.
    The terms "breakpoint" and "crossover" are used synonymously. A
    breakpoint's position is the index of the breakpoint's second amino acid in
    the parent alignment. This definition allows the breakpoint position to
    nicely slice the parent alignment into blocks. For example, let the
    position of the k-1th, kth, and k+1th breakpoints be b_k_1, b_k, b_k__1,
    respectively. Then for parent sequence p, the block before the breakpoint
    is p[b_k_1:b_k] and the block after is p[b_k:b_k__1]. Valid breakpoints
    contain candidate overhangs, defined below.

    Overhangs are the DNA sticky ends used in a Golden Gate reaction. In this
    module, each overhang consists of a positional shift and DNA sequence. The
    former is overhang's position relative to the first base pair in the left
    codon of the breakpoint (codon index b_k - 1).

    Example: Suppose the amino acids at indices b_k-1 and b_k at 'I' and 'T'
        for all parents in the parent alignment and codon_options contains
        {'M': ('ATG,), 'T': ('ACC')}}. Then for each parent p, the only valid
        codon sequence for p[b_k-1:b_k+1] is ATGACC, and the overhangs are
        (0, 'ATGA'), (1, 'TGAC'), (2, 'GACC'). This breakpoint will be
        represented in the breakpoints dictionary as
        {b_k: [(0, 'ATGA'), (1, 'TGAC'), (2, 'GACC')]}.

    Certain adjacent breakpoints will result in the same <M>, e.g. if alignment
    positions i and i+1 only consist of {'A'} and {'C'} respectively, the
    libraries of breakpoints (..., i, ...) and (..., i+1, ...) will have the
    same <M> if all other breakpoints are the same. Therefore, this function
    groups the redundant breakpoints and returns a mapping from the original
    breakpoint index to the nonredundant group index.

    Args:
        parent_aln: parent alignment for breakpoint calculation.
        codon_options: Amino acid to available codon mapping. Used in
            Golden Gate site design and library sequence design. Change
            this to include or exclude certain codons based on codon
            optimization schemes, reassigned codons, etc.
        start_overhang: Positional shift and nucleotide sequence of Golden
            Gate site for vector insertion at start of sequence. Not
            factored into calculations if None.
        end_overhang: Positional shift and nucleotide sequence of Golden
            Gate site for vector insertion at end of sequence. Not factored
            into calculations if None.

    Returns:
        breakpoints: Mapping from breakpoint position to valid breakpoint
            overhangs.
        bp_to_group: Mapping from breakpoint index to index of nonredundant
            breakpoint groups.
    """

    # Check that codon_options contain all AAs in parent alignment.
    print('hello')
    print(parent_aln)
    aa_alphabet: list[tuple[str, ...]] = set().union(*parent_aln)
    print(aa_alphabet)
    assert all(aa in codon_options for aa in aa_alphabet)

    # Sets of codons for each amino acid at an alignment position.
    aln_cdns = [{codon_options[aa] for aa in pos_aas}
                for pos_aas in parent_aln]

    breakpoints = {}

    if start_overhang is not None:
        breakpoints[0] = [start_overhang]
        internal_bp_start = 1
    else:
        internal_bp_start = 0

    # Search possible Golden Gate sites by iterating over adjacent codons.
    for bp, (cdns1, cdns2) in enumerate(zip(aln_cdns, aln_cdns[1:]), 1):

        # Can't do Golden Gate at a site with gaps.
        # TODO: Do further Golden Gate validation.
        if '---' in cdns1 or '---' in cdns2:
            continue

        # Get the [len 1], [len 2], and [len 3] patterns for the two sites
        # as list of lists.
        pos1_bins = _get_valid_patterns(cdns1, reverse=True)
        pos2_bins = _get_valid_patterns(cdns2)

        overhangs = []

        # Combine the pattern sets to get all the len 4 overhangs. The
        # pos1 list is reversed so that the len 3 patterns go with the
        # pos2 len 1 patterns, len 2 with len 2, etc.
        pattern_iter = zip(reversed(pos1_bins), pos2_bins)
        for pos, (pat_set1, pat_set2) in enumerate(pattern_iter):
            for pat1, pat2 in product(pat_set1, pat_set2):
                oh = (pos, pat1 + pat2)
                overhangs.append(oh)

        if not overhangs:
            # No valid overhangs, skip.
            continue

        breakpoints[bp] = overhangs

    if end_overhang is not None:
        breakpoints[len(aln_cdns)] = [end_overhang]
        internal_bp_end = len(breakpoints) - 1
    else:
        internal_bp_end = len(breakpoints)

    # TODO: Check if there's enough breakpoints for the run params.
    # Group the redundant breakpoints.
    breakpoint_groups = []

    # Iterate over the internal (not added start or end) breakpoints.
    bps = iter(sorted(breakpoints)[internal_bp_start:internal_bp_end])
    curr_bps = [next(bps)]
    for bp in bps:
        # If bp is not adjacent to last or has more than one amino acid, it
        # belongs to a new group.
        if bp != curr_bps[-1] + 1 or len(set(parent_aln[bp-1])) != 1:
            breakpoint_groups.append(curr_bps)
            curr_bps = [bp]
        else:
            curr_bps.append(bp)
    breakpoint_groups.append(curr_bps)

    # Get the mapping from each breakpoint index to the breakpoint's group
    # index.
    bp_to_group = {}
    for group_index, bp_group in enumerate(breakpoint_groups):
        for bp in bp_group:
            bp_to_group[bp] = group_index

    return breakpoints, bp_to_group


class _Edge(NamedTuple):
    """Edge between nodes in RASPP graph.

    Public Attributes:
        e_diff: Difference in energy between out_node and in_node.
        in_node: Input node.
        out_node: Output_node.
    """
    e_diff: float
    in_node: '_Node'
    out_node: '_Node'


@dataclass(repr=False)
class _Node:
    """Node in RASPP graph.

    Public Attributes:
        col: Column where this node is located.
        index: Position of breakpoint represented by this node.
        in_edges: Edges that point to this node.
        out_edges: Mapping from edge.index to edge for each edge with this node
            as in_node.

    Public Methods:
        fill_out: Fill out edges from this node to next column.
    """
    col: int
    index: int
    overhangs: Optional[list[tuple[int, str]]]
    in_edges: list[_Edge] = field(default_factory=list)
    out_edges: list[_Edge] = field(default_factory=list)

    @property
    def breakpoint(self):
        return {self.index: self.overhangs}

    def fill_out(
        self,
        next_col: list['_Node'],
        energy_func: EnergyFunction
    ) -> None:
        """Fill out edges from this node to next column.

        Args:
            next_col: Next column in graph.
            energy_func: Used to calculate energy between graph nodes.
        """

        def add_edge(energy, target_node):
            """Add an edge from self to target_node."""
            edge = _Edge(e, self, target_node)
            # self.out_edges[target_node.index] = edge
            self.out_edges.append(edge)
            target_node.in_edges.append(edge)

        # Next_col must have sorted indices.
        assert all(n1.index < n2.index for n1, n2
                   in zip(next_col, next_col[1:]))

        # Iterate on column, calculating energy and adding edges as needed.
        col_iter = iter(next_col)

        # Find a node such that the index is greater than this node's.
        # TODO: account for maxBL and minBL
        curr_node = next(col_iter)
        while curr_node.index <= self.index:
            curr_node = next(col_iter)

        # Get the energy of the first valid node in the column, make edge.
        e = energy_func.block_energy(self.index, curr_node.index)
        add_edge(e, curr_node)

        # Start iterating over breakpoint columns starting at next index.
        start_index = curr_node.index + 1

        # Get the next node. If this is a StopIteration, we are at the end of
        # the column, so we can return.
        try:
            curr_node = next(col_iter)
        except StopIteration:
            return

        for i in range(start_index, next_col[-1].index):
            # Calculate energy difference using previous energy and energy
            # difference between previous and current indicies.
            e += energy_func.increment_block_energy(self.index, i)

            # If we're at curr_node's index, we can add an edge to the node.
            if i == curr_node.index:
                add_edge(e, curr_node)
                curr_node = next(col_iter)

        # Add the last edge. Doing it here avoids the StopIteration.
        add_edge(e, curr_node)

    def __hash__(self):
        return hash(repr(self))

    def __repr__(self):
        in_edges = [e.in_node.index for e in self.in_edges]
        out_edges = [e.out_node.index for e in self.out_edges]
        ret = f'Node({self.col}, {self.index}, {self.breakpoint}, ' \
            f'{in_edges=}, {out_edges=})'
        return ret


if __name__ == '__main__':
    import sys

    loc = '../../tests/bgl3_sample/'
    pa = ParentAlignment.from_fasta(loc+'bgl3_sequences.fasta')
    pdb = PDBStructure.from_pdb_file(loc+'1GNX.pdb')
    pa.pdb_structure = pdb

    sys.exit()

    '''
    loc = '../tests/bgl3_sample/truncated/'
    pa = ParentAlignment.from_fasta(loc+'trunc.fasta')
    pdb = PDBStructure.from_pdb_file(loc+'trunc.pdb')
    '''
    loc = '../../tests/bgl3_sample/'
    pa = ParentAlignment.from_fasta(loc+'bgl3_sequences.fasta')
    pdb = PDBStructure.from_pdb_file(loc+'1GNX.pdb')
    pa.pdb_structure = pdb

    vector_overhangs = [(0, 'TATG'), (3, 'TGAG')]
    n = 4
    raspp = RASPP(pa, n, vector_overhangs[0], vector_overhangs[1])

    '''
    aln_sets = [set(a) for a in pa]
    breakpoint_groups = []
    bps = iter(sorted(raspp.breakpoints)[1:-1])
    curr_bps = [next(bps)]
    for bp in bps:
        print(bp, aln_sets[bp-1], aln_sets[bp])
        if bp != curr_bps[-1] + 1 or len(aln_sets[bp-1]) != 1:
            breakpoint_groups.append(curr_bps)
            curr_bps = [bp]
        else:
            curr_bps.append(bp)
    breakpoint_groups.append(curr_bps)
    print(breakpoint_groups)
    '''

    minBL, maxBL = 70, 105
    # minBL, maxBL = 30, 105
    import time

    # varied
    s = time.time()
    libs = raspp.vary_m_proxy(minBL, maxBL)
    '''
    print('\nvary_m_proxy')
    for rl in sorted(libs, key=lambda x: x.energy):
        print(rl.energy, list(rl.breakpoints), rl.min_block_len,
              rl.max_block_len, rl.average_m)
    '''
    print(time.time() - s)

    # all
    print('\nall_libraries')
    s = time.time()
    all_libs = raspp.all_libraries(minBL, maxBL)
    print('here')
    print(len(all_libs))
    for i, rl in enumerate(sorted(all_libs, key=lambda x: x.energy)):
        print(i, rl.energy, list(rl.breakpoints), rl.min_block_len,
              rl.max_block_len, rl.average_m)
    print(time.time() - s)
    1/0

    import matplotlib.pyplot as plt
    lib_me = set((rl.average_m, rl.energy) for rl in libs)
    alib_me = set((rl.average_m, rl.energy) for rl in all_libs)
    lib_plot = lib_me - alib_me
    alib_plot = alib_me - lib_me
    both_plot = lib_me & alib_me
    print(lib_plot)
    print(alib_plot)
    print(both_plot)

    all_breakpoints = [sorted(rl.breakpoints) for rl in all_libs]
    for rl in libs:
        if sorted(rl.breakpoints) not in all_breakpoints:
            print(sorted(rl.breakpoints))

    if lib_plot:
        plt.scatter(*zip(*lib_plot), label='varied')
    if alib_plot:
        plt.scatter(*zip(*alib_plot), label='all')
    plt.scatter(*zip(*both_plot), label='both')
    plt.xlabel('<M>')
    plt.ylabel('<E>')
    plt.legend()
    plt.show()

    """
    rl = min(all_libs, key=lambda x: x.energy)
    print(rl)
    print(rl.min_block_len, rl.max_block_len)
    """

    '''
    for minBL, maxBL in [(1, 8), (2, 5), (None, None)]:
        all_libs = raspp.all_libraries(minBL, maxBL)
        print(minBL, maxBL, [sorted(lib.breakpoints) for lib in all_libs])
    '''
    '''
    raspp.eval_all_bps()
    raspp.min_bps()

    N = len(pa.aligned_sequences[0].seq)
    opt_time = 0.0
    opt_single_time = 0.0
    naive_time = 0.0
    for bps in combinations(range(1, 9), 5):
        m = average_m(pa, (0,) + bps)
        print(bps, m)
    '''



#!/usr/bin/env python

import sys
import os

import yaml

USAGE_INFO = 'Usage: ggsr --help\n' \
             '       ggsr <config filename>'

dirname = os.path.split(os.path.abspath(os.path.realpath(sys.argv[0])))[0]

HELP_INFO = 'ggsr is a program for designing SCHEMA-RASPP libraries with ' \
            'Golden Gate Assembly.\n\n' \
            'The following procedure will design a library that can ' \
            'be assembled with BsaI Golden Gate Assembly into a vector with ' \
            'BsaI overhangs: \n' \
            '    1. Make a new directory.\n' \
            '    2. Copy a fasta file with your parent names and sequences ' \
            'into the directory.\n' \
            '    3. Optionally copy a .pdb file with structure information ' \
            'related to your parent sequences into the directory.\n' \
            f'    4. Inside the directory, run "cp {dirname}/' \
            'run_config.yaml ."\n' \
            '    5. Edit this configuration file according to your library' \
            ' requirements.\n' \
            '    6. Run "nohup ggsr run_config.yaml &> ggsr.log &".\n\n' \
            f'Alternatively, run "cp -r {dirname} <new directory name>",' \
            ' edit the stepx.py scripts, and run them individually to ' \
            'get more flexibility.\n' \
            'View https://github.com/RomeroLab/SCHEMA-library-design/wiki ' \
            'for more information.'

if len(sys.argv) != 2:
    print(USAGE_INFO)
    sys.exit()

if sys.argv[1] == '--help':
    print(HELP_INFO)
    sys.exit()

with open(sys.argv[1]) as f:
    config_map = yaml.safe_load(f)

excluded_steps = config_map['excluded_steps']

if 1 not in excluded_steps:
    sys.argv = [sys.argv[0], config_map['seq_fn'], config_map['pdb_fn'],
                config_map['out_prefix']]
    run_args = [str(c)+' ' for c in sys.argv]
    run_args[0] = 'step1.py '
    print('Running:', ''.join(run_args))
    exec(open(dirname + '/step1.py').read())
    print('Step1 done\n')

if 2 not in excluded_steps:
    sys.argv = [sys.argv[0],
                config_map['num_bl'], config_map['minBL'], config_map['maxBL'],
                config_map['start_oh_seq'], config_map['start_oh_pos'],
                config_map['end_oh_seq'], config_map['end_oh_pos'],
                config_map['out_prefix']+'_AA.fasta',
                config_map['out_prefix']+'_contacts.json',
                config_map['libraries_fn']]
    run_args = [str(c)+' ' for c in sys.argv]
    run_args[0] = 'step2.py '
    print('Running:', ''.join(run_args))
    exec(open(dirname + '/step2.py').read())
    print('Step2 done\n')

if 3 not in excluded_steps:
    sys.argv = [sys.argv[0], config_map['libraries_fn'],
                config_map['gg_prob_threshold'], config_map['chosen_lib_fn']]
    run_args = [str(c)+' ' for c in sys.argv]
    run_args[0] = 'step3.py '
    print('Running:', ''.join(run_args))
    exec(open(dirname + '/step3.py').read())
    print('Step3 done\n')

if 4 not in excluded_steps:
    sys.argv = [sys.argv[0], config_map['out_prefix']+'_AA.fasta',
                config_map['out_prefix']+'_CDN.fasta',
                config_map['chosen_lib_fn'], config_map['frags_order_fn']]
    run_args = [str(c)+' ' for c in sys.argv]
    run_args[0] = 'step4.py '
    print('Running:', ''.join(run_args))
    exec(open(dirname + '/step4.py').read())
    print('Step4 done')



# parent_alignment.py

"""Alignments of parental sequences for combinatorial protein libraries.

This module provides the definition of :class:`ggrecomb.ParentAlignment`, which
represents an alignment of parental protein sequences used in ggrecomb library
creation. The module can be used directly to use accessory functions for
calculating identity between sequences, querying the BLAST web interface, or
selecting additional parent sequences from a list of candidates.

"""

from collections import defaultdict
from itertools import combinations
from io import StringIO
import json
import os
from subprocess import CalledProcessError
from subprocess import run
import time
from typing import Optional
from urllib.parse import urlencode
from urllib.request import urlopen
from urllib.request import Request

from Bio import Entrez
from Bio import pairwise2
from Bio import SeqIO
from Bio.Blast.NCBIWWW import _parse_qblast_ref_page
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

import ggrecomb
from ggrecomb import PDBStructure


def calc_identity(sr1: SeqRecord, sr2: SeqRecord) \
        -> float:
    """Calculate the BLAST identity between two sequences.

    Note:
        "BLAST identity" is defined as the number of matches in a pairwise
        alignment divided by the length of alignment. This is different from
        the "traditional" definition of identity: number of matches divided by
        the minimum of the sequence lengths.

    Parameters:
        sr1: SeqRecord of first sequence to compare.
        sr2: SeqRecord of second sequence to compare.

    Returns:
        The BLAST identity between the two sequences.

    """
    s1 = str(sr1.seq)
    s2 = str(sr2.seq)
    aln_score = pairwise2.align.globalxx(s1, s2, score_only=True)
    aln_len = len(s1) + len(s2) - aln_score
    return aln_score / aln_len


def query_blast(
    query_seq: str,
    database: str = 'refseq_protein',
    maximum_hits: int = 10000
) -> SeqIO.FastaIO.FastaIterator:
    """Gets sequences from query_seq BLAST hits.

    Direct copy of Peter Cock's Bio.Blast.NCBIWWW.qblast, with simplifications
    and added progress print-outs.

    Args:
        query_seq: Query sequence for BLAST search.
        database: BLAST database to search. Must be either 'refseq_protein' or
            'pdbaa'
        maximum_hits: Maximum number of hits allowed to be returned. Affects
            runtime: reducing this number is faster but you risk missing good
            candidate sequences.

    Returns:
        FASTA generator for sequence hits.

    Raises:
        ValueError: If database is not "refseq_protein" or "pdbaa".
    """

    NCBI_BLAST_URL = "https://blast.ncbi.nlm.nih.gov/Blast.cgi"

    # Only refseq_protein and pdbaa databases supported.
    if database not in ('refseq_protein', 'pdbaa'):
        raise ValueError('database must be "refseq_protein" or "pdbaa"')

    # Initiate BLAST search.
    params = [
        ('PROGRAM', 'blastp'),
        ('DATABASE', database),
        ('QUERY', query_seq),
        ('HITLIST_SIZE', maximum_hits),
        ("CMD", "Put"),
    ]
    message = urlencode(params).encode()
    request = Request(NCBI_BLAST_URL, message, {"User-Agent":
                                                "BiopythonClient"})
    # TODO: Change from BiopythonClient.
    handle = urlopen(request)

    # Get BLAST Request ID.
    rid, _ = _parse_qblast_ref_page(handle)

    # Setup BLAST job checking request.
    params = [
        ('RID', rid),
        ('FORMAT_TYPE', 'XML'),
        ('CMD', 'Get'),
    ]
    message = urlencode(params).encode()
    request = Request(NCBI_BLAST_URL, message, {"User-Agent":
                                                "BiopythonClient"})

    # Query BLAST once a minute until job is done.
    previous = 0.0
    delay = 20.0  # seconds
    while True:
        current = time.time()
        wait = previous + delay - current
        if wait > 0:
            time.sleep(wait)
            previous = current + wait
        else:
            previous = current
        # delay by at least 60 seconds only if running the request against the
        # public NCBI API
        if delay < 60:
            # Wasn't a quick return, must wait at least a minute
            delay = 60

        # print('Checking for job completion...')
        handle = urlopen(request)
        results = handle.read().decode()

        # Can see an "\n\n" page while results are in progress,
        # if so just wait a bit longer...
        if results == "\n\n":
            continue
        # XML results don't have the Status tag when finished
        if "Status=" not in results:
            break
        i = results.index("Status=")
        j = results.index("\n", i)
        status = results[i + len("Status="): j].strip()
        if status.upper() == "READY":
            break

    # print('Job complete.')

    # Request to get hit accessions from search results.
    params = [
        ('RID', rid),
        ('RESULTS_FILE', 'on'),
        ('FORMAT_TYPE', 'CSV'),
        ('DESCRIPTIONS', maximum_hits),
        ('FORMAT_OBJECT', 'Alignment'),
        ('DOWNLOAD_TEMPL', 'Results'),
        ('QUERY_INDEX', '0'),
        ('CMD', 'Get'),
    ]
    message = urlencode(params).encode()
    request = Request(NCBI_BLAST_URL, message, {"User-Agent":
                                                "BiopythonClient"})
    handle = urlopen(request)

    # Unpack accessions into candidate_accessions list.
    handle.readline()  # headers
    candidate_accessions = []
    for line in handle:
        fields = line.decode().strip().split(',')
        if len(fields) <= 1:
            continue
        # Split by ',' unfortunately splits last field in two.
        accession = fields[-1].replace('"', '').replace(')', '')
        candidate_accessions.append(accession)

    # TODO: query for new sequences on demand (generator)

    # Get accession sequences from NCBI.
    acc_str = ','.join(candidate_accessions)
    Entrez.email = 'bbremer@wisc.edu'  # this probably needs to change
    fasta_str = Entrez.efetch(db='protein', id=acc_str, rettype='fasta')

    # print('returning')

    return SeqIO.parse(fasta_str, 'fasta')


class _NoNewNodeError(RuntimeError):
    """Raised when no node is added to the tree."""
    pass


class _TreeNode:
    """Represents a set of candidate sequences.

    Helper class for choose_candidates function.

    Attributes:
        cands: List of candidates SeqRecords.
        max_cc_diff: Maximum out of abs(% identity - target_identity) between
        each pair of candidates.
        max_pc_diff: Maximum out of abs(% identity - target_identity) of each
            parent and each candidate.
        children: Child nodes of this node.
        parent: Parent node of this node.
    """
    def __init__(
        self,
        cands: SeqRecord = [],
        max_cc_diff: float = 0.0,
        max_pc_diff: float = 0.0,
        parent: '_TreeNode' = None
    ) -> None:
        """Initialize node instance with precomputed identies."""
        self.cands = cands
        self.max_cc_diff = max_cc_diff
        self.max_pc_diff = max_pc_diff
        self.children: list['_TreeNode'] = []
        if parent is not None:
            self.parent = parent

    def new_cand(self, cand: SeqRecord, new_pc_diff: float,
                 diff_thresh: float, target_identity: float) -> '_TreeNode':
        """Create new child node with candidates self.cands + [cand].

        Args:
            cand: New candidate sequence.
            new_pc_diff: Maximum out of abs(% identity - target_identity)
                between cand and each parent. This is assumed to be greater
                than the "pc_diff" of anything in self.cands, by virtue of
                pre-sorting based on the pc_diff.
            diff_thresh: Current known optimum for a valid set of candidates.
                If the potentially new set has a higher diff than this value,
                we can skip creation because this node and all its children
                are non-optimal.
            target_identity: Ideal cross-wise identity between all sequences in
                the concatenation set of parents and selected candidates.
        """
        if self.cands:
            # Calculate the identity diff between the new candidate and each
            # candidate in self.cands.
            new_cc_diff = max(abs(target_identity - calc_identity(cand, x))
                              for x in self.cands)
            # Choose maximum between old cc_diff and new_cc_diff.
            new_cc_diff = max(self.max_cc_diff, new_cc_diff)
        else:
            new_cc_diff = 0.0  # only one candidate in new node
        if max(new_cc_diff, new_pc_diff) >= diff_thresh:
            # don't create a new node because it is already non-optimal
            raise _NoNewNodeError
            # return None  # probably change b/c EP2 Item 20
        new_cands = self.cands + [cand]
        return _TreeNode(new_cands, new_cc_diff, new_pc_diff, self)

    def add_cand(self, cand: SeqRecord, new_pc_diff: float,
                 diff_thresh: float, target_identity: float) -> None:
        """new_cand new_cand for adding non-leaf nodes."""
        try:
            new_node = self.new_cand(cand, new_pc_diff, diff_thresh,
                                     target_identity)
        except _NoNewNodeError:
            return
        if new_node is not None:
            self.children.append(new_node)

    def delete(self) -> None:
        """Delete node from tree."""
        self.parent.children.remove(self)

    def __repr__(self) -> str:
        """Return CSV string of the node's candidate ids."""
        return ','.join([c.id for c in self.cands])


class _Tree:
    """Data structure for finding the ideal set of candidate sequences.

    Helper class for choose_candidates function.

    Attributes:
        base: Node representing the empty set of candidates.
        num_seq: Number of sequences to choose. Depth of tree is num_seq + 1.
        target_identity: Ideal cross-wise identity between all sequences in
            the concatenation set of parents and selected candidates.
        best_leaf: The TreeNode with the current best set of candidates.
        best_diff: The maximum abs(% identity - target_identity) out of each
            pair in best_leaf.cands + parents.
    """
    def __init__(self, num_seqs: int, target_identity: float):
        """Tree for finding <num_seqs> candidates with <target_identity>."""
        self.base = _TreeNode()
        self.num_seqs = num_seqs
        self.target_identity = target_identity
        self.best_diff = 1.0

    def add_cand(self, cand: SeqRecord, new_pc_diff: float) -> Optional[str]:
        """Add cand to candidate sets in tree and add new nodes if needed.

        Args:
            cand: New candidate sequence.
            new_pc_diff: Maximum out of abs(% identity - target_identity)
                between cand and each parent. This is assumed to be greater
                than the "pc_diff" of anything in self.cands, by virtue of
                pre-sorting based on the pc_diff.
        """
        # Breadth-first traversal over tree, adding new cand to each node.
        stack = [self.base]
        while stack:
            curr_node = stack.pop()

            # Prune tree if curr_node's max identity difference is greater
            # than the current known best.
            curr_max = max(curr_node.max_cc_diff, curr_node.max_pc_diff)
            if curr_max > self.best_diff:
                curr_node.delete()
                continue

            # Leaf node case: if new node will have num_seqs cands, then we can
            # evaluate it directly and don't need to add it to tree.
            if len(curr_node.cands) == self.num_seqs - 1:
                try:
                    leaf = curr_node.new_cand(cand, new_pc_diff,
                                              self.best_diff,
                                              self.target_identity)
                except _NoNewNodeError:
                    continue
                # leaf max diff is smaller than self.best_diff, new best!
                self.best_leaf = leaf
                self.best_diff = max(leaf.max_cc_diff, leaf.max_pc_diff)

                # Every future leaf will have equal or greater max_pc_diff,
                # so this leaf must be the best one and we can stop.
                if leaf.max_pc_diff >= leaf.max_cc_diff:
                    return 'best found'

                continue

            # Non-leaf node case.
            stack += curr_node.children  # comes first b/c don't want dup cands
            curr_node.add_cand(cand, new_pc_diff, self.best_diff,
                               self.target_identity)
        return None

    def shape(self) -> list[int]:
        """Number of nodes at each level in tree."""
        shape: dict[int, int] = defaultdict(int)
        stack = [self.base]
        while stack:
            curr_node = stack.pop()
            shape[len(curr_node.cands)] += 1
            stack += curr_node.children
        return [shape[i] for i, _ in enumerate(shape)]


def choose_candidates(
    candidate_sequences: list[SeqRecord],
    existing_parents: list[SeqRecord],
    num_additional: int,
    desired_identity: float
) -> list[SeqRecord]:
    """Choose the ideal set of candidate sequences.

    Ideal set is defined as the set of candidates with the minimum max_diff,
    where max_diff is the maximum cross-wise abs(% identity - desired_identity)
    between each pair of sequences in the concatenation of the set and parents.

    Args:
        candidate_sequences: Sequences able to be selected.
        existing_parents: Parent sequences in the library already.
        num_additional: Number of candidate sequences to choose.
        desired_identity: Ideal cross-wise identity between all sequences in
            the concatenation set of parents and selected candidates.

    Returns:
        Ideal set of candidate SeqRecords.

    Raises:
        ValueError: if nonpositive num_additional provided, desired_identity
            not bet 0.0 and 1.0 (noninclusive), or less candidate_sequences
            than num_additional provided.
    """

    # Parameter checking.
    if num_additional < 1:
        raise ValueError('num_additional must be positive.')
    if not 0.0 < desired_identity < 1.0:
        raise ValueError('desired_identity must be between 0.0 and 1.0, '
                         'exclusive.')
    if len(candidate_sequences) < num_additional:
        raise ValueError('Insufficient number of candidates provided.')
    if not existing_parents:
        raise ValueError('existing_parents must not be empty.')
    # TODO: Handle empty existing_parents as valid? I.e. choose starting from
    # nothing. (Version 0.2.0)

    # For each candidate, find the maximum identity difference with each
    # parent.

    cand_diffs = []
    # TODO: logging (Version 0.2.0)
    # print('Number of candidates:', len(candidate_sequences))
    # print('Calculating pc_diff and sorted')
    for i, cand in enumerate(candidate_sequences):
        # print(i, '\r', end='')
        max_diff = max(abs(desired_identity - calc_identity(cand, x))
                       for x in existing_parents)
        cand_diffs.append((cand, max_diff))
    # print()

    # Sort candidates by difference to enable short-circuiting. sorted is fast
    # enough, but this could be faster with heapq.
    sorted_cand_diffs = list(sorted(cand_diffs, key=lambda x: x[1]))

    # Construct Tree and find the best set. In the worst case this might take
    # awhile.
    # TODO: Bound the time this requires?
    # print('Constructing tree.')
    tree = _Tree(num_additional, desired_identity)
    for i, (cand, pc_diff) in enumerate(sorted_cand_diffs):
        # print(i, '\r', end='')
        if pc_diff > tree.best_diff:
            break
        ret = tree.add_cand(cand, pc_diff)
        if ret == 'best found':
            break
    # print()
    return tree.best_leaf.cands


def web_muscle(records: list[SeqRecord]) -> str:
    """Align sequences using MUSCLE on EMBL-EBI Web Services.

    Parameters:
        records: The SeqRecords holding the sequences to align.

    Return:
        The alignment in a FASTA format string. It can be read manually or
            with "SeqIO.parse(StringIO(aln_str), 'fasta')".

    """
    email = 'bbremer@wisc.edu'
    base_url = 'https://www.ebi.ac.uk/Tools/services/rest/muscle/'

    seqs_f = StringIO('')
    SeqIO.write(records, seqs_f, 'fasta')
    seqs_f.seek(0)
    seqs_str = seqs_f.read()

    # Submit job.
    params = {
        'email': email,
        'format': 'fasta',
        'sequence': seqs_str,
    }
    message = urlencode(params).encode()
    url = base_url + 'run'
    request = Request(url, message)
    jobid = urlopen(request).read().decode()

    # Wait for completion.
    url = base_url + 'status/' + jobid
    response = urlopen(url).read().decode()
    while response == 'RUNNING':
        time.sleep(1)
        response = urlopen(url).read().decode()
    if response != 'FINISHED':
        url = base_url + 'result/' + jobid + '/error'
        error = urlopen(url).read().decode()
        raise RuntimeError(f"MUSCLE status response: {repr(response)}, "
                           f"error message: {error}")

    # Get resulting alignment.
    url = base_url + 'result/' + jobid + '/aln-fasta'
    aln_str = urlopen(url).read().decode()
    return aln_str


def local_muscle(records: list[SeqRecord]) -> str:
    """Align sequences with MUSCLE and set to aligned_sequences.

    WARNING: Experimental in the current version.
    """
    # print('running alignment')

    IN_FN = 'temp_muscle_input.fasta'
    OUT_FN = 'temp_muscle_output.fasta'

    # make file for MUSCLE input
    # TODO: Use tempfile package.
    SeqIO.write(records, IN_FN, 'fasta')

    # run MUSCLE
    try:
        run(f'muscle -in {IN_FN} -out {OUT_FN}', shell=True, check=True)
    except CalledProcessError:
        # print('Something is wrong with MUSCLE call. Is MUSCLE installed?')
        raise
    finally:
        os.remove(IN_FN)

    with open(OUT_FN) as f:
        aln_str = f.read()
    os.remove(OUT_FN)
    return aln_str


class _ParentSequences:
    """
    Parent protein sequences for recombinant library design.

    This class sets up the data needed to run recombinant design algorithms,
    e.g. intaking and aligning parental sequences, aligning the parents, and
    finding a PDB structure or additional parents. Instances of this class are
    passed into functions further down the ggrecomb pipeline.

    Note:
        The first sequence (records[0]) has special importance, as it's used to
        find PDB structures or additional parents. For parent sequence sets
        with high enough identity (~60%), this should generally not be an
        issue, but you may get different results by changing the order of the
        sequences.

    Parameters:
        records: Parental amino acid sequences for ggrecomb
            calculations.
        auto_align: If True, the records are aligned upon initialization.
        prealigned: If True, the records are already aligned upon
            initialization. auto_align and prealigned must not both be
            True.
        pdb_structure: Protein Data Bank structure that represents the three-
            dimensional structure of the aligned parent sequences. Will be
            automatically renumbered if either auto_align or prealigned is
            True.

    Attributes:
        records (list[SeqRecord]): Parental amino acid sequences for
            ggrecomb calculations. `BioPython SeqRecords
            <https://biopython.org/wiki/SeqRecord>`_ contain sequence metadata
            such as the name of the sequence. The (unaligned) ith sequence may
            be obtained as a Python string with
            "str(parents[i].seq)". Changing this attribute will delete the
            alignment attribute.
        alignment (list[tuple[str, ...]]): Alignment of records, with the
            aligned strings in the same order as the source records. Present if
            and only if the instance is aligned. To set this attribute, call
            the align or set_alignment methods, which will renumber the
            pdb_structure attribute if present.
        p0_aligned (str): records[0] aligned, calculated from alignment.
            Present if and only if instance is aligned.
        pdb_structure (PDBStructure): Protein Data Bank structure that
            represents the three-dimensional structure of the aligned parent
            sequences.

    Examples:
        To start, you need a FASTA file with at least one parent sequence. For
        these examples, this file is "bgl3_sequence.fasta", which contains the
        HIS-tagged beta-gluocosidase sequence associated with the Protein Data
        Bank (PDB) entry "1GNX".

        The easiest, but slowest way to use this class is to let it handle
        everything through web services. For example, the following script
        will build a six-parents ParentAlignment with roughly 70% identity
        between the parents and the PDB structure closest to the parents.

        >>> from ggrecomb import ParentSequences
        >>> parents = ParentSequences.from_fasta('bgl3_sequence.fasta')
        >>> parents.obtain_seqs(6, 0.7)  # BLAST takes about 10 minutes.
        >>> parents.align()  # MUSCLE takes about a minute.
        >>> parents.get_PDB()  # BLAST takes about 10 minutes.
        >>> len(parents.records)
        6
        >>> parents.p0_aligned  # doctest: +ELLIPSIS
        'MHHHHHHMVPAAQQTAMA...RTGVLPTA-------'
        >>> parents.p0_aligned == parents.pdb_structure.renumbering_seq
        True


        You can skip the slow web queries if you already have a FASTA with the
        aligned parents and the PDB structure you want to use:

        >>> from ggrecomb import ParentSequences
        >>> from ggrecomb import PDBStructure
        >>> pdb = PDBStructure.from_pdb_file('1GNX.pdb')
        >>> parents = ParentSequences.from_fasta(
        ...     'bgl3_aligned_parents.fasta',
        ...     pdb_structure=pdb,
        ...     prealigned=True
        ... )
        >>> len(parents.records)
        6
        >>> parents.p0_aligned  # doctest: +ELLIPSIS
        'MHHHHHHMVPAAQQTAMA...RTGVLPTA-------'
        >>> parents.p0_aligned == parents.pdb_structure.renumbering_seq
        True

        You can also save or load a ParentAlignment as a JSON:

        >>> from ggrecomb import ParentSequences
        >>> from ggrecomb import PDBStructure
        >>> pdb = PDBStructure.from_pdb_file('1GNX.pdb')
        >>> parents = ParentSequences.from_fasta(
        ...     'bgl3_aligned_parents.fasta',
        ...     pdb_structure=pdb,
        ...     prealigned=True
        ... )
        >>> parents_json = parents.to_json()
        >>> with open('parents.json', 'w') as f:
        ...     f.write(parents_json)
        ...
        339501
        >>> with open('parents.json', 'r') as f:
        ...     parents_json2 = f.read()
        ...
        >>> parents2 = ParentSequences.from_json(parents_json2)
        >>> # parents and parents2 are the same.
        >>> parents.alignment == parents2.alignment
        True

    """

    def __init__(
        self,
        records: list[SeqRecord],
        pdb_structure: Optional['PDBStructure'] = None,
        auto_align: bool = False,
        prealigned: bool = False
    ) -> None:
        # Input checking.
        if len(records) < 1:
            raise ValueError('records must not be empty.')
        if auto_align and prealigned:
            raise ValueError('auto_align and prealigned must not both be '
                             'True.')

        self.records = records

        self.pdb_structure: 'PDBStructure'
        if pdb_structure is not None:
            self.pdb_structure = pdb_structure

        if auto_align:
            self.align()
        if prealigned:
            self.new_alignment([str(rec.seq) for rec in records])

    def __setattr__(self, name, value) -> None:
        if name == 'alignment':
            raise AttributeError('The alignment attribute cannot be set '
                                 'directly. Use the align or new_alignment '
                                 'methods.')

        super().__setattr__(name, value)

        if name == 'records' and hasattr(self, '_alignment'):
            # If sequences is changed, alignment is out of date, delete.
            del self._alignment

        if (name == '_alignment' and hasattr(self, 'pdb_structure')) or \
           (name == 'pdb_structure' and hasattr(self, '_alignment')):
            # Renumber existing pdb_structure if alignment changes or if
            # renumber new pdb_structure if self is aligned.
            self.pdb_structure.renumber(self.p0_aligned)

    @property
    def alignment(self):
        try:
            return self._alignment
        except AttributeError:
            raise AttributeError('ParentSequences instance is not aligned.')

    @property
    def p0_aligned(self):
        return ''.join([aminos[0] for aminos in self.alignment])

    def new_alignment(self, aligned_sequences: list[str]):
        """Add aligned sequences from records to instance.

        Sets the alignment attribute to the input alignments.

        Parameters:
            aligned_sequences: Aligned sequences in the same order as the
                records attribute.

        Raises:
            ValueError: If provided sequences do not match the sequences in the
                records attribute.

        """
        if len(aligned_sequences) != len(self.records):
            raise ValueError('The number of aligned sequences provided must be'
                             ' the same as the number of records.')
        aln_len = len(aligned_sequences[0])
        if any(len(aln_seq) != aln_len for aln_seq in aligned_sequences[1:]):
            raise ValueError('All sequences must be the same length if '
                             'prealigned.')
        for sr, aln_seq in zip(self.records, aligned_sequences):
            if str(sr.seq).replace('-', '') != aln_seq.replace('-', ''):
                raise ValueError('aligned_sequences is does not match '
                                 'self.records.')
        self._alignment = list(zip(*aligned_sequences))

    def align(self, run_locally: bool = False) -> None:
        """Use the MUSCLE web service to align the records.

        Sets the alignment attribute to the resulting alignment. The
        run_locally attribute is currently experimental and probably shouldn't
        be used.

        """
        aln_str = None
        if run_locally:
            try:
                aln_str = local_muscle(self.records)
            except CalledProcessError:
                print('muscle command not available, falling back to MUSCLE '
                      'web service.')

        # Try to use the MUSCLE web API five times.
        for _ in range(5):
            try:
                aln_str = web_muscle(self.records)
                break
            except RuntimeError as e:
                print(e)

        if aln_str is None:
            raise RuntimeError('MUSCLE web API did not work.')

        aln_f = StringIO(aln_str)
        out_SRs = {SR.name: SR for SR in SeqIO.parse(aln_f, 'fasta')}
        reordered_srs = [out_SRs[sr.name] for sr in self.records]

        # This renumbers the self.pdb_structure if present.
        aligned_sequences = [str(sr.seq) for sr in reordered_srs]
        self._alignment = list(zip(*aligned_sequences))

    def obtain_seqs(
        self,
        num_final_sequences: int,
        desired_identity: Optional[float] = None
    ) -> None:
        """Adds new sequences with BLAST.

        Parameters:
            num_final_sequences: Number of desired parent sequences. After
                call, instance should have len(sequences) equal this.
            desired_identity: Desired percent identity of new sequences
                compared to first parent sequence. Default: if there are
                multiple parent sequences already, this value is taken as the
                average identity between the first parent and other sequences.
                Otherwise, 70% identity.

        Raises:
            ValueError: if num_final_sequences is not greater than
                len(self.records) or if desired_identity is a float and not
                between 0.0 and 1.0, exclusive.

        """
        if len(self.records) >= num_final_sequences:
            len_seqs = len(self.records)
            raise ValueError(f'Requested {num_final_sequences} total sequences'
                             f' but there are already {len_seqs} sequences.')

        # This results in two calls, but we want to validate before BLAST.
        desired_identity = self._check_identity(desired_identity)

        query_sr = self.records[0]
        query_seq = str(query_sr.seq)
        candidate_seqs = list(query_blast(query_seq))
        # query_blast gives iter but list should be small enough

        self.add_from_candidates(candidate_seqs, num_final_sequences,
                                 desired_identity)

    def add_from_candidates(
        self,
        candidate_sequences: list[SeqRecord],
        num_final_sequences: int,
        desired_identity: float = None
    ) -> None:
        """Add new parent sequences from list of candidates.

        Finds the set of sequences in candidate_sequences that gives the
        smallest maximum difference between desired_identity and the calculated
        identity between any two sequences in the set.

        Parameters:
            candidate_sequences: Sequences to choose from.
            num_final_sequences: Number of desired parent sequences. After
                call, instance should have len(sequences) equal this.
            desired_identity: Desired identity of new sequences
                compared to first parent sequence. Default: if there are
                multiple parent sequences already, this value is taken as the
                average identity between the first parent and other sequences.
                Otherwise, 70% identity.

        Raises:
            ValueError: if num_final_sequences is not greater than
                len(self.records) or if desired_identity is a float and not
                between 0.0 and 1.0, exclusive.

        """

        num_preexisting_parents = len(self.records)
        num_additional = num_final_sequences - num_preexisting_parents
        if num_additional <= 0:
            return
        if len(candidate_sequences) < num_additional:
            raise ValueError('candidate_sequences is not large enough to '
                             f'add {num_additional} sequences.')
        desired_identity = self._check_identity(desired_identity)

        best_cands = choose_candidates(
            candidate_sequences,
            self.records,
            num_additional,
            desired_identity
        )
        self.records += best_cands

    def get_PDB(self) -> None:
        """Construct from ParentAlignment using BLAST and PDB.

        The best structure is found by using BLAST to download candidate PDB
        sequences, then the sequence with the largest minimum identity to the
        parents is selected and set to the parent_alignment attribute.

        Parameters:
            parent_aln: Parent alignment used to query the PDB. Note that
                parent_aln[0] is used in the query and PDBStructure alignment.

        Raises:
            ValueError: If no matching PDB structure could be found.

        """
        query_str = str(self.records[0].seq)

        pdb_srs = list(query_blast(query_str, 'pdbaa', 100))

        # Find the PDB struct with largest minimum identity to the parents.
        pdb_min_map = {}
        for pdb_sr in pdb_srs:
            min_iden = min(calc_identity(parent, pdb_sr) for parent
                           in self.records)
            pdb_min_map[pdb_sr.id] = min_iden

        try:
            best_id = max(pdb_min_map, key=lambda pdbid: pdb_min_map[pdbid])
        except ValueError:
            raise RuntimeError('No best PDB could be found.')

        # Get the pdb_structure from rcsb.
        _, acc, chain = best_id.split('|')
        url = 'https://files.rcsb.org/view/' + acc + '.pdb'
        with urlopen(url) as f:
            pdb_structure = PDBStructure.from_pdb_file(f, chain=chain)

        self.pdb_structure = pdb_structure

    @classmethod
    def from_fasta(cls, fasta_fn: str, **kwargs) -> '_ParentSequences':
        """Contruct instance from FASTA file.

        Parameters:
            fasta_fn: filename of FASTA file, including relative path.
            **kwargs: Additional keyword args for __init__. For example, you
                can specify "auto_align=True" in this constructor.

        Return:
            ParentsSequences instance constructed from input FASTA file.

        """
        seqs = list(SeqIO.parse(fasta_fn, 'fasta'))
        return cls(seqs, **kwargs)

    def to_json(self) -> str:
        """Convert instance to a JSON-formatted string.

        Return:
            Instance converted to a JSON string.

        """

        records = [[str(sr.seq), sr.id, sr.name, sr.description] for sr
                   in self.records]

        # TODO: probably add a version?
        out_list: list = [records]

        try:
            out_list.append(self._alignment)
        except AttributeError:
            out_list.append(None)

        try:
            out_list.append(self.pdb_structure.to_json())
        except AttributeError:
            out_list.append(None)

        return json.dumps(out_list)

    @classmethod
    def from_json(cls, in_json: str):
        """Construct instance from JSON.

        Parameters:
            in_json: JSON-formatted string representing a ParentsSequences.

        Return:
            ParentsSequences instance created from in_json.

        """
        records, alignment, pdb = json.loads(in_json)

        seq_records = []
        for sr_list in records:
            sr_seq, sr_id, sr_name, sr_desc = sr_list
            seq = Seq(sr_seq)
            sr = SeqRecord(seq, id=sr_id, name=sr_name, description=sr_desc)
            seq_records.append(sr)

        if pdb is not None:
            pdb = PDBStructure.from_json(pdb)

        new_instance = cls(seq_records, pdb)

        if alignment is not None:
            # This will renumber the pdb_structure but that's okay.
            alignment = [tuple(ele) for ele in alignment]
            new_instance._alignment = alignment

        return new_instance

    def _check_identity(self, desired_identity: Optional[float]) -> float:
        """Validate desired_identity, set to records average if None."""
        if desired_identity is None:
            # Follow default behavior: 0.7 if there's only one sequence.
            # Average cross-wise identity of the sequences otherwise.
            if len(self.records) == 1:
                desired_identity = 0.7
            else:
                # calculate average pairwise identity of parents
                identities = [calc_identity(sr1, sr2) for sr1, sr2
                              in combinations(self.records, 2)]
                desired_identity = sum(identities) / len(identities)
        if not 0.0 < desired_identity < 1.0:
            raise ValueError('desired_identity must be between 0.0 and 1.0, '
                             'exclusive.')
        return desired_identity



# pdb_structure.py

"""Parse and manipulate Protein Data Bank structure files.

This module provides the definition of :class:`ggrecomb.PDBStructure` and the
accessory classes :class:`AminoAcid` and :class:`Atom` that represent the
eponymous entities with a PDB structure.

Parsing and modifying protein structure data is necessary for recombinant
library design because SCHEMA energy calculations require the distance between
amino acids within the protein.

These structures are obtained from the Protein Data Bank at
https://www.rcsb.org and loaded as Python objects. Nearly all information is
discarded except for ATOM lines, which specify data about atoms within the
structure.

The most confusing part about PDB structure manipulation is the renumbering of
atoms to match SCHEMA-RASPP parent alignment. Relieving user confusion about
this process is a primary goal of this module. Note that atom/residue indexing
in PDB files begins at 1, while Python's indexing starts at 0. For consistency
with the rest of the package, these classes index the residue number starting
from 0. As a result, PDB file reading and writing PDB must convert between
indexing. For example, the "ALA" lines in the pdb file below indicate that the
15th amino acid in the structure is alanine. When read with this module, this
alanine will be labeled with an index of 14. This is consistent with
sequence number: if pdb_seq is the Python String representing the amino acid
sequence of the structure , this residue would be pdb_seq[14].

PDB files generally have this structure (example structure 1GNX)::

 ...<other structure data>...
 ATOM      1  N   ALA A  15      -1.611  17.176  10.792  1.00 36.46           N
 ATOM      2  CA  ALA A  15      -1.871  18.610  11.107  1.00 36.85           C
 ATOM      3  C   ALA A  15      -2.021  18.795  12.611  1.00 36.41           C
 ATOM      4  O   ALA A  15      -2.983  18.321  13.215  1.00 38.36           O
 ATOM      5  CB  ALA A  15      -3.131  19.081  10.392  1.00 35.10           C
 ATOM      6  N   LEU A  16      -1.064  19.496  13.206  1.00 34.22           N
 ATOM      7  CA  LEU A  16      -1.061  19.738  14.642  1.00 29.97           C
 ATOM      8  C   LEU A  16      -1.711  21.073  14.992  1.00 30.29           C
 ATOM      9  O   LEU A  16      -1.462  22.089  14.341  1.00 30.34           O
 ATOM     10  CB  LEU A  16       0.380  19.716  15.152  1.00 26.33           C
 ATOM     11  CG  LEU A  16       1.228  18.548  14.639  1.00 24.12           C
 ATOM     12  CD1 LEU A  16       2.681  18.761  15.026  1.00 22.75           C
 ATOM     13  CD2 LEU A  16       0.698  17.230  15.195  1.00 23.37           C
 ATOM     14  N   THR A  17      -2.541  21.066  16.028  1.00 30.68           N
 ATOM     15  CA  THR A  17      -3.217  22.278  16.472  1.00 29.15           C
 ATOM     16  C   THR A  17      -2.576  22.784  17.756  1.00 28.03           C
 ATOM     17  O   THR A  17      -2.337  22.012  18.683  1.00 28.04           O
 ATOM     18  CB  THR A  17      -4.716  22.014  16.733  1.00 30.86           C
 ATOM     19  OG1 THR A  17      -5.357  21.666  15.501  1.00 32.64           O
 ATOM     20  CG2 THR A  17      -5.385  23.246  17.319  1.00 30.50           C
 ...<other atoms>...
 ...<other structure data>...

"""

from dataclasses import dataclass
from functools import cached_property
from itertools import combinations
import json
from pathlib import Path
from typing import Optional, TextIO, Union

from Bio import pairwise2
from Bio.SeqUtils import seq1
import numpy as np
from scipy.spatial import distance


@dataclass
class Atom:
    """Atom within a PDB structure.

    This class is basically a one-to-one conversion of 'ATOM' lines in `PDB
    files <https://www.cgl.ucsf.edu/chimera/docs/UsersGuide/tutorials/pdbintro
    .html>`_. The primary constructor is Atom.from_line() to allow for easy
    parsing of PDB files. For ggrecomb, the important attributes are res_name,
    res_index, x, y, and z.

    Initalization parameters and attributes are equivalent for this class.

    Parameters:
        serial_num: Atom serial number.
        name: Atom name.
        alt_loc: Alternate location indicator.
        res_name: Residue name.
        chain: Chain identifier.
        res_index: Residue index within the structure..
        insertion_code: Code for insertions of residues.
        x: X orthogonal Angstrom coordinate.
        y: Y orthogonal Angstrom coordinate.
        z: Z orthogonal Angstrom coordinate.
        occup: Occupancy.
        temp: Temperature factor.
        segment: Segment identifier.
        element: Element symbol.
        charge: Charge.

    """
    serial_num: int
    name: str
    alt_loc: str
    res_name: str
    chain: str
    res_index: int
    insertion_code: str
    x: float
    y: float
    z: float
    occup: float
    temp: float
    segment: str
    element: str
    charge: str

    @classmethod
    def from_line(cls, line: str) -> 'Atom':
        """Construct Atom from PDB ATOM line.

        Parameters:
            line: ATOM line from PDB file.

        Returns:
            Atom constructed from line.

        Raises:
            ValueError: If line has more than one newline, a newline not at the
                end of the line, first characters that are not "ATOM", or more
                than one occurence of ATOM, or if an Atom cannot otherwise
                be constructed from the line.

        """
        num_newlines = line.count('\n')
        if num_newlines > 1 or (num_newlines == 1 and line[-1] != '\n'):
            raise ValueError('More than one line passed into constructor.')

        if line[:4] != 'ATOM' or line.count('ATOM') != 1:
            raise ValueError('Invalid ATOM line passed into ATOM.from_line():'
                             f' {line}')

        try:
            serial_num = int(line[6:11].strip())
            name = line[12:16].strip()
            alt_loc = line[16]
            res_name = line[17:20].strip()
            chain = line[21]
            res_index = int(line[22:26].strip()) - 1  # switch to 0 indexing
            insertion_code = line[26]
            x = float(line[30:38].strip())
            y = float(line[38:46].strip())
            z = float(line[46:54].strip())
            occup = float(line[54:60].strip())
            temp = float(line[60:66].strip())
            segment = line[72:76].strip()
            element = line[76:78].strip()
            charge = line[78:80].strip()
        except ValueError:
            raise ValueError('Invalid ATOM line passed into ATOM.from_line():'
                             f' {line}')

        return cls(serial_num, name, alt_loc, res_name, chain, res_index,
                   insertion_code, x, y, z, occup, temp, segment, element,
                   charge)

    def to_line(self) -> str:
        """Convert to PDB file ATOM line.

        Returns:
            str of ATOM line in PDB file format.

        """
        line = list('ATOM') + [' ' for _ in range(75)]
        line[6:11] = str(self.serial_num).rjust(5)
        # Atom names are complicated, see the "misalignment" section of the
        # PDB format link, found in the Atom class docstring. We take a
        # technically incorrect, but more common approach.
        line[13:16] = self.name.ljust(4)
        line[16] = self.alt_loc
        line[17:20] = self.res_name.rjust(3)
        line[21] = self.chain
        line[22:26] = str(self.res_index+1).rjust(4)  # switch to 1 indexing
        line[26] = self.insertion_code
        line[30:38] = f'{self.x:.3f}'.rjust(8)
        line[38:46] = f'{self.y:.3f}'.rjust(8)
        line[46:54] = f'{self.z:.3f}'.rjust(8)
        line[54:60] = f'{self.occup:.2f}'.rjust(6)
        line[60:66] = f'{self.temp:.2f}'.rjust(6)
        line[72:76] = self.segment.ljust(4)
        line[76:78] = self.element.rjust(2)
        line[78:80] = self.charge.rjust(2)
        return ''.join(line)


class AminoAcid:
    """Amino acid within a PDB structure.

    This class is a collection of :class:`Atom` objects that make up a residue
    in a :class:`ggrecomb.PDBStructure`, plus additional functions for ggrecomb
    calculation.

    Parameters:
        atoms: Atoms within the AminoAcid. All Atom objects in list must have
            the same res_name and res_index. res_index must be indexed from 0,
            not 1, but this should have happened on Atom initialization.

    Attributes:
        atoms: The residue's atoms.
        name: Name of the residue, e.g. ALA, LEU.
        index: Index of the residue within the structure.
        orig_index: Original struct_index. Not set unless the AminoAcid is
            renumbered.
        letter: name converted to one letter code.
        coords: x, y, and z coordinates of each atom in a Nx3 numpy ndarray,
            where N is the number of atoms.

    Raises:
        ValueError: During initialization, if the list is empty or atoms do not
            all have the same res_name and res_index.
        AttributeError: If AminoAcid has not been renumbered and orig_index is
            accessed.

    """
    def __init__(self, atoms: list[Atom]):
        if not atoms:
            raise ValueError('atoms must not be an empty list.')

        self.atoms = atoms

        res_name = atoms[0].res_name
        if any(a.res_name != res_name for a in atoms):
            raise ValueError('Atoms do not all have the same res_name.')
        self.name = res_name

        res_index = atoms[0].res_index
        if any(a.res_index != res_index for a in atoms):
            raise ValueError('Atoms do not all have the same res_index.')
        self.index = res_index

    @property
    def letter(self) -> str:
        three_letters = self.name[0] + self.name[1:].lower()
        return seq1(three_letters)

    @cached_property
    def coords(self) -> np.ndarray:  # shape (# atoms, 3)
        return np.asarray([(a.x, a.y, a.z) for a in self.atoms])

    def renumber(self, new_index: int) -> None:
        """Change the structure index of the amino acid and its atoms."""
        self.orig_index = self.index
        self.index = new_index
        for a in self.atoms:
            a.res_index = new_index

    def derenumber(self) -> None:
        """Revert changes made by the renumber method."""
        old_index = self.orig_index
        del self.orig_index
        self.index = old_index
        for a in self.atoms:
            a.res_index = old_index

    @classmethod
    def from_lines(cls, lines: list[str]) -> 'AminoAcid':
        """Construct AminoAcid from ATOM lines in a PDB file.

        Parameters:
            lines: ATOM lines in a PDB file.

        Returns:
            AminoAcid constructed from lines.

        """
        atoms = [Atom.from_line(line) for line in lines]
        return cls(atoms)

    def to_lines(self) -> list[str]:
        """Convert to section of PDB ATOM lines.

        Returns:
            PDB format line for each of the AminoAcid's atoms.

        Note:
            This method does not save the orig_index attribute.

        Example:
         >>> from ggrecomb.pdb_structure import AminoAcid
         >>> pdb_fn = 'tests/fixtures/bgl3_full/1GNX.pdb'
         >>> in_lines = []
         >>> with open(pdb_fn) as f:
         >>>     for line in f:
         >>>         if line[:4] != 'ATOM':
         >>>             continue
         >>>         res_index = int(line[22:26].strip())
         >>>         # 1GNX starts at 15, see pdb_structure module docs.
         >>>         if res_index == 15:
         >>>             in_lines.append(line)
         >>>         else:
         >>>             break
         >>> aa = AminoAcid.from_lines(in_lines)
         >>> out_lines = aa1.to_lines()
         >>> all(i == o for i, o in zip(in_lines, out_lines))
         True

        """
        return [a.to_line() for a in self.atoms]

    @classmethod
    def from_json(cls, in_json: str) -> 'AminoAcid':
        """JSON from_lines wrapper that includes orig_index.

        Parameters:
            in_json: JSON-formatted input string of a list that contains a list
                of ATOM lines and orig_index, the latter of which may be null.

        Returns:
            AminoAcid constucted from input JSON string.

        """
        lines, orig_index = json.loads(in_json)
        obj = cls.from_lines(lines)
        if orig_index is not None:
            obj.orig_index = orig_index
        return obj

    def to_json(self) -> str:
        """JSON to_lines wrapper that includes orig_index.

        Returns:
            JSON string representing the AminoAcid.

        Example:
         >>> from ggrecomb import _PDBStructure
         >>> from ggrecomb.pdb_structure import AminoAcid
         >>> pdb_fn = 'tests/fixtures/bgl3_full/1GNX.pdb'
         >>> pdb = PDBStructure.from_pdb_file(pdb_fn)
         >>> aa1 = pdb.amino_acids[0]
         >>> json_str = aa1.to_json()
         >>> aa2 = AminoAcid.from_json(json_str)
         >>> aa1.name == aa2.name
         True
         >>> aa1.index == aa2.index
         True

        """
        orig_index: Optional[int]
        try:
            orig_index = self.orig_index
        except AttributeError:
            orig_index = None
        return json.dumps([self.to_lines(), orig_index])


class _PDBStructure:
    """Structure from the Protein Data Bank. Used in downstream calculations.

    This class is a Python representation of a PDB file, e.g. `1GNX.pdb
    <https://files.rcsb.org/download/1GNX.pdb>`_, which are downloaded from
    https://www.rcsb.org. This download may be done automatically with the
    :meth:`ggrecomb.ParentAlignment.get_PDB` method or manually. In the latter
    case, PDBStructure object must also be created manually, e.g. using the
    :meth:`from_pdb_file` class method.

    Note:
        There is no guarantee that the amino_acids attribute is ordered
        by the index of the amino acid.

    See :mod:`ggrecomb.pdb_structure` for more information on the details of
    this class.

    Parameters:
        amino_acids: Amino acids in the PDB structure.
        unrenumbered_amino_acids: The amino acids not included in renumbering.
            None if PDBStructure is not renumbered. Must be None if
            renumbering_seq is None.
        renumbering_seq: Sequence used to renumber structure. None if
            PDBStructure is not renumbered. Must not None if
            unrenumbered_amino_acids is None.

    Attributes:
        amino_acids: (list[AminoAcid]): Residues in the PDB structure.
        unrenumbered_amino_acids (list[AminoAcid]): Original amino_acids not
            included in the renumbering. Present if and only if structure is
            renumbered.
        renumbering_seq (str): Sequence that was used to renumber structure.
            Present if and only if structure is renumbered.
        seq (str): Amino acid sequence of structure, including gaps based on
            the indices of the amino_acids. Note that this attribute will
            change if renumbering occurs.
        contacts (list[tuple[int, int]]): Indices of contacting residues, where
            a pair of residues are contacting if the shortest distance between
            them is less than 4.5 angstroms.

    Raises:
        ValueError: During initialization, if input amino_acids have non-unique
            indicies, unrenumbered_amino_acids is an empty list,
            renumbering_seq does not match amino_acids, or one but not both of
            unrenumbered_amino_acids and renumbering_seq is None.
        AttributeError: If PDBStructure is not renumbered and
            unrenumbered_amino_acids or renumbering_seq is accessed.

    """

    def __init__(
        self,
        amino_acids: list[AminoAcid],
        unrenumbered_amino_acids: Optional[list[AminoAcid]] = None,
        renumbering_seq: Optional[str] = None
    ):
        # Ensure that all amino_acid have unique indices.
        indices = set()
        for aa in amino_acids:
            index = aa.index
            if index in indices:
                raise ValueError('AminoAcid objects in a PDBStructure must '
                                 'have unique indices.')
            indices.add(index)

        self.amino_acids = amino_acids
        self._contacts: list[tuple[int, int]]

        # If PDBStructure is already aligned, both unrenumberd_amino_acids and
        # renumbering_seq must both be not None.
        if unrenumbered_amino_acids is not None \
           and renumbering_seq is not None:
            if any(aa.letter != renumbering_seq[aa.index] for aa
                   in amino_acids):
                raise ValueError(f'renumbering seq "{renumbering_seq}" is '
                                 'invalid with input amino_acids.')

            self.unrenumbered_amino_acids = unrenumbered_amino_acids
            self.renumbering_seq = renumbering_seq
        elif unrenumbered_amino_acids is not None \
                or renumbering_seq is not None:
            raise ValueError('Unrenumbered_amino_acids and renumbering_seq '
                             'must both be None if one is None.')

    @property
    def seq(self) -> str:
        seq_dict = {aa.index: aa.letter for aa in self.amino_acids}
        return ''.join([seq_dict.get(i, '-') for i in range(max(seq_dict)+1)])

    @property
    def contacts(self) -> list[tuple[int, int]]:
        try:
            # Cached contacts.
            return self._contacts
        except AttributeError:
            contacts = []
            aa_coords = {aa.index: aa.coords for aa in self.amino_acids}
            combos = combinations(aa_coords, 2)
            for ii, j in combos:
                a1 = aa_coords[ii]
                a2 = aa_coords[j]
                d = distance.cdist(a1, a2)
                if np.any(d < 4.5):
                    contacts.append((ii, j))
                    contacts.append((j, ii))
            self._contacts = contacts
            return contacts

    def renumber(self, p1_aligned: str) -> None:
        """Renumber pdb structure to match ParentAlignment.

        Parameters:
            p1_aligned: Sequence to align to. Usually the first parent from
                a ParentAlignment.

        """
        # If already renumbered, derenumber to get all amino_acids back.
        if hasattr(self, 'unrenumbered_amino_acids'):
            self.derenumber()

        # Will put each AminoAcid into one of these.
        renumbered_amino_acids = []
        unrenumbered_amino_acids = []

        # Order self.amino_acids.
        index_to_amino_acids = {aa.index: aa for aa in self.amino_acids}
        ordered_amino_acids = [index_to_amino_acids[i] for i
                               in sorted(index_to_amino_acids)]

        # Align p1_aligned and the PDBStructure, using '.' for gaps.
        pdb_seq = ''.join([aa.letter for aa in ordered_amino_acids])
        aln = pairwise2.align.globalxx(p1_aligned, pdb_seq, gap_char='.',
                                       one_alignment_only=True)[0]
        aligned_p1_seq = aln.seqA
        aligned_pdb_seq = aln.seqB

        # Simultaneously iterate through ordered amino_acids and the
        # alignment. Every time we see a non-gap character in aligned_pdb_seq,
        # the next element in pdb_iter is the AminoAcid corresponding to that
        # character. If p1 character at the same position in the alignment is
        # also not a gap, we renumber the AminoAcid to the parent_index and
        # include it.
        pdb_iter = iter(ordered_amino_acids)
        parent_index = 0  # Tracks the index of the p1.
        alignment_iter = zip(aligned_p1_seq, aligned_pdb_seq)
        for i, (par_aa, pdb_aa) in enumerate(alignment_iter):
            if pdb_aa != '.':
                candidate_aa = next(pdb_iter)
                if par_aa != '.':
                    # Aligned residue. Renumber to parent and include.
                    candidate_aa.renumber(parent_index)
                    renumbered_amino_acids.append(candidate_aa)
                    parent_index += 1
                else:
                    # In pdb but not p1, include in unrenumbered_amino_acids.
                    # print(f'warning: residue {i} in pdb but not parent.')
                    unrenumbered_amino_acids.append(candidate_aa)
            elif par_aa != '.':
                # Not in pdb, but in p1, increment parent_index.
                parent_index += 1

        self.amino_acids = renumbered_amino_acids
        self.unrenumbered_amino_acids = unrenumbered_amino_acids
        self.renumbering_seq = p1_aligned
        if hasattr(self, '_contacts'):
            del self._contacts

    def derenumber(self) -> None:
        """Revert any changes made by the renumber method back to original.

        Raises:
            AttributeError: If PDBStructure is not renumbered.

        """
        try:
            unrenumbered_amino_acids = self.unrenumbered_amino_acids
            del self.unrenumbered_amino_acids
        except AttributeError:
            raise AttributeError('PDBStructure has not been renumbered.')

        for aa in self.amino_acids:
            aa.derenumber()
        amino_acids = self.amino_acids + unrenumbered_amino_acids
        self.amino_acids = amino_acids
        del self.renumbering_seq
        if hasattr(self, '_contacts'):
            del self._contacts

    @classmethod
    def from_pdb_file(
        cls,
        f: Union[str, Path, TextIO],
        chain: str = 'A'
    ) -> '_PDBStructure':
        """Construct from PDB file without renumbering.

        Parameters:
            f: file-like PDB structure.
            chain: Chain to include in constructed object.

        Returns:
            PDBStructure initialized from PDB file.

        """
        if isinstance(f, str) or isinstance(f, Path):
            f = open(f)
            close_later = True
        else:
            close_later = False

        amino_acids = []
        curr_atoms: list[Atom] = []
        for line in f:
            if line[:4] not in (b'ATOM', 'ATOM'):
                continue
            if isinstance(line, bytes):
                line = line.decode()
            atom = Atom.from_line(line)
            if atom.chain != chain:
                continue
            if curr_atoms and curr_atoms[0].res_index != atom.res_index:
                aa = AminoAcid(curr_atoms)
                amino_acids.append(aa)
                curr_atoms = [atom]
            else:
                curr_atoms.append(atom)

        if close_later:
            f.close()

        if curr_atoms:
            aa = AminoAcid(curr_atoms)
            amino_acids.append(aa)
        return cls(amino_acids)

    def to_json(self) -> str:
        """Convert structure to JSON string.

        Returns:
            JSON string representing the PDBStructure.

        """
        aa_jsons = [aa.to_json() for aa in self.amino_acids]
        unre_jsons: Optional[list[str]]
        try:
            unre_jsons = [aa.to_json() for aa in self.unrenumbered_amino_acids]
        except AttributeError:
            unre_jsons = None
        try:
            renum_seq: Optional[str] = self.renumbering_seq
        except AttributeError:
            renum_seq = None
        return json.dumps([aa_jsons, unre_jsons, renum_seq])

    @classmethod
    def from_json(cls, in_json: str) -> '_PDBStructure':
        """Construct from JSON string.

        Parameters:
            in_json: JSON string representing a PDBStructure instance.

        Returns:
            PDBStructure constructed from JSON string.

        Example::

        >>> aminos = [...]  # list of AminoAcids
        >>> s1 = PDBStructure(aminos)
        >>> json_str = s1.to_json()
        >>> s2 = PDBStructure.from_json(in_str)
        >>> # Show s1 and s2 are the same.
        >>> for aa1, aa2 in zip(s1.amino_acids, s2.amino_acids):
        >>>     assert aa1.name == aa2.name
        >>>     assert aa1.index == aa2.index

        """
        aa_jsons, unre_jsons, renumbering_seq = json.loads(in_json)
        amino_acids = [AminoAcid.from_json(aa_json) for aa_json in aa_jsons]
        unrenumbered: Optional[list[AminoAcid]]
        if unre_jsons is None:
            unrenumbered = None
        else:
            unrenumbered = [AminoAcid.from_json(aa_json) for aa_json
                            in unre_jsons]
        return cls(amino_acids, unrenumbered, renumbering_seq)
